# -*- coding: utf-8 -*-
import ast
from typing import TYPE_CHECKING
from collections import deque

if TYPE_CHECKING:
    from typing import Deque, Dict, Optional, Tuple, Union


def compute_lineno_to_stmt_mapping(node: 'Union[str, ast.AST]') -> 'Dict[int, ast.stmt]':
    line_to_stmt_map = {}
    if isinstance(node, str):
        node = ast.parse(node)
    q: Deque[Tuple[ast.AST, Optional[ast.stmt]]] = deque([(node, None)])
    while len(q) > 0:
        node, stmt_parent = q.pop()
        if isinstance(node, ast.stmt):
            stmt_parent = node
        if stmt_parent is not None and hasattr(node, 'lineno'):
            line_to_stmt_map[node.lineno] = stmt_parent
        for field, value in ast.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        q.appendleft((item, stmt_parent))
            elif isinstance(value, ast.AST):
                q.appendleft((value, stmt_parent))
    return line_to_stmt_map



# -*- coding: utf-8 -*-
import functools

from nbsafety.run_mode import SafetyRunMode


def return_arg_at_index(index, logger):
    def return_arg_func(exc, *args, **kwargs):
        logger.warning('Exception occurred: %s' % exc)
        if args is not None and len(args) > index:
            return args[index]
        return None
    return return_arg_func


def return_val(val, logger):
    def return_val_func(exc, *args, **kwargs):
        logger.warning('Exception occurred: %s' % exc)
        return val
    return return_val_func


def on_exception_default_to(recovery_func):
    def make_wrapper(original_func):
        @functools.wraps(original_func)
        def wrapped_func(*args, **kwargs):
            try:
                return original_func(*args, **kwargs)
            except Exception as e:
                if SafetyRunMode.get() == SafetyRunMode.DEVELOP:
                    raise e
                else:
                    return recovery_func(e, *args, **kwargs)
        return wrapped_func
    return make_wrapper



# -*- coding: utf-8 -*-
from .symbol_edges import get_assignment_lval_and_rval_symbol_refs, get_symbol_edges
from .attr_symbols import AttrSubSymbolChain, CallPoint, get_attrsub_symbol_chain
from .live_refs import compute_live_dead_symbol_refs
from .utils import compute_call_chain_live_symbols, get_symbols_for_references, stmt_contains_lval



# -*- coding: utf-8 -*-
import logging

from .utils import assert_bool, make_safety_fixture

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
_safety_fixture, _safety_state, run_cell_ = make_safety_fixture(setup_cells=['from nbsafety.utils import DotDict'])
run_cell = run_cell_


def stale_detected():
    return _safety_state[0].test_and_clear_detected_flag()


def assert_detected(msg=''):
    assert_bool(stale_detected(), msg=msg)


def assert_not_detected(msg=''):
    assert_bool(not stale_detected(), msg=msg)


def lookup_symbol(val):
    safety = _safety_state[0]
    val_id = id(val)
    if val_id not in safety.aliases:
        return None
    alias_set = safety.aliases[val_id]
    if len(alias_set) == 0:
        return None
    return next(iter(alias_set))


def test_basic():
    run_cell('d = DotDict()')
    run_cell('d.x = DotDict()')
    run_cell('d.y = DotDict()')
    run_cell('d.x.a = 5')
    # run_cell('d.x.b = 6')
    run_cell('logging.info(d.x.a)')
    assert_not_detected()
    run_cell('logging.info(d.x)')
    assert_not_detected()
    run_cell('d.y.a = 7')
    # run_cell('d.y.b = 8')
    run_cell('logging.info(d.y.a)')
    assert_not_detected()
    run_cell('logging.info(d.y)')
    assert_not_detected()
    run_cell('x = d.x.a + 9')
    run_cell('y = d.y.a + 9')
    run_cell('d.y.a = 9')
    run_cell('logging.info(x)')
    assert_not_detected('`x` independent of changed `d.y.a`')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on changed `d.y.a`')
    run_cell('logging.info(d.x.a)')
    assert_not_detected()
    run_cell('logging.info(d.x)')
    assert_not_detected()
    run_cell('logging.info(d)')
    assert_not_detected()
    run_cell('d.y = 10')
    run_cell('logging.info(x)')
    assert_not_detected('`x` independent of changed `d.y`')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on changed `d.y`')


def test_nested_readable_name():
    run_cell('d = DotDict()')
    run_cell('d.x = DotDict()')
    run_cell('d.x.a = 5')
    run_cell('d.x.b = 6')
    d_x_a = lookup_symbol(5)
    assert d_x_a.readable_name == 'd.x.a'
    d_x_b = lookup_symbol(6)
    assert d_x_b.readable_name == 'd.x.b'
    run_cell('d.y = DotDict()')
    run_cell('d.y.a = 7')
    run_cell('d.y.b = 8')
    d_y_a = lookup_symbol(7)
    assert d_y_a.readable_name == 'd.y.a'
    d_y_b = lookup_symbol(8)
    assert d_y_b.readable_name == 'd.y.b'



# -*- coding: utf-8 -*-
from enum import Enum

from nbsafety.utils import fast


EMIT_EVENT = '_X5ix_NBSAFETY_EVT_EMIT'


class TraceEvent(Enum):
    before_stmt = 'before_stmt'
    after_stmt = 'after_stmt'

    attribute = 'attribute'
    subscript = 'subscript'
    after_attrsub_chain = 'after_attrsub_chain'

    before_arg_list = 'enter_arg_list'
    after_arg_list = 'exit_arg_list'
    argument = 'argument'

    before_literal = 'before_literal'
    after_literal = 'after_literal'

    line = 'line'
    call = 'call'
    return_ = 'return'
    exception = 'exception'

    # these are included for completeness but will probably not be used
    c_call = 'c_call'
    c_return = 'c_return'
    c_exception = 'c_exception'

    def __str__(self):
        return self.value

    def to_ast(self):
        return fast.Constant(self.value)



# -*- coding: utf-8 -*-
import os
import sys
from typing import TYPE_CHECKING, cast

from IPython import get_ipython
import pytest

from nbsafety.run_mode import SafetyRunMode
from nbsafety.safety import NotebookSafety

if TYPE_CHECKING:
    from typing import Any, List, Optional, Tuple


def should_skip_known_failing(reason='this test tests unimpled functionality'):
    return {
        'condition': os.environ.get('SHOULD_SKIP_KNOWN_FAILING', True),
        'reason': reason
    }


skipif_known_failing = pytest.mark.skipif(**should_skip_known_failing())


def assert_bool(val, msg=''):
    assert val, str(msg)


# Reset dependency graph before each test to prevent unexpected stale dependency
def make_safety_fixture(**kwargs) -> 'Tuple[Any, List[Optional[NotebookSafety]], Any]':
    os.environ[SafetyRunMode.DEVELOP.value] = '1'
    safety_state: List[Optional[NotebookSafety]] = [None]

    def run_cell(code, ignore_exceptions=False):
        get_ipython().run_cell_magic(safety_state[0].cell_magic_name, None, code)
        try:
            if not ignore_exceptions and getattr(sys, 'last_value', None) is not None:
                raise sys.last_value
        finally:
            sys.last_value = None

    store_history = kwargs.pop('store_history', False)
    test_context = kwargs.pop('test_context', True)
    setup_cells = kwargs.pop('setup_cells', [])

    @pytest.fixture(autouse=True)
    def init_or_reset_dependency_graph():
        safety_state[0] = NotebookSafety(
            cell_magic_name='_SAFETY_CELL_MAGIC',
            store_history=store_history,
            test_context=test_context,
            **kwargs
        )
        run_cell('import sys')
        run_cell('sys.path.append("./test")')
        run_cell('import logging')
        for setup_cell in setup_cells:
            run_cell(setup_cell)
        yield  # yield to execution of the actual test
        # ensure each test didn't give failures during ast transformation
        exc = safety_state[0].set_ast_transformer_raised(None)
        if exc is not None:
            raise exc
        get_ipython().reset()  # reset ipython state

    return init_or_reset_dependency_graph, safety_state, run_cell



# -*- coding: utf-8 -*-

# need PYTHONPATH="." for this to work
import builtins
try:
    import test
except ImportError:
    import sys
    sys.path.append(".")
    import test

if __name__ == '__main__':
    setattr(builtins, '__exit_zero', True)
    for name, mod in test.__dict__.items():
        # weird; need to reimport these for some reason
        import sys
        import ipytest
        if name.startswith('test_') and isinstance(mod, type(sys)):
            if ipytest.run(*sys.argv[1:], filename=mod.__file__, return_exit_code=True) != 0:
                import builtins
                setattr(builtins, '__exit_zero', False)
    # Totally bizarre; we lose all our variables
    # Extreme hack to keep the zero exit status around
    import sys
    import builtins
    sys.exit(0) if getattr(builtins, '__exit_zero') else sys.exit(1)



import ast
import copy
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Dict, List, Set


class StatementMapper(ast.NodeVisitor):
    def __init__(self, line_to_stmt_map: 'Dict[int, ast.stmt]', id_map: 'Dict[int, ast.AST]'):
        self.line_to_stmt_map = line_to_stmt_map
        self.id_map = id_map
        self.traversal: 'List[ast.AST]' = []

    def __call__(self, node: 'ast.AST') -> 'Dict[int, ast.AST]':
        # for some bizarre reason we need to visit once to clear empty nodes apparently
        self.visit(node)
        self.traversal.clear()
        
        self.visit(node)
        orig_traversal = self.traversal
        self.traversal = []
        self.visit(copy.deepcopy(node))
        copy_traversal = self.traversal
        orig_to_copy_mapping = {}
        for no, nc in zip(orig_traversal, copy_traversal):
            orig_to_copy_mapping[id(no)] = nc
            self.id_map[id(nc)] = nc
            if isinstance(nc, ast.stmt):
                self.line_to_stmt_map[nc.lineno] = nc
                # workaround for python >= 3.8 wherein function calls seem
                # to yield trace frames that use the lineno of the first decorator
                for decorator in getattr(nc, 'decorator_list', []):
                    self.line_to_stmt_map[decorator.lineno] = nc
        return orig_to_copy_mapping

    def visit(self, node):
        self.traversal.append(node)
        for name, field in ast.iter_fields(node):
            if isinstance(field, ast.AST):
                self.visit(field)
            elif isinstance(field, list):
                for inner_node in field:
                    if isinstance(inner_node, ast.AST):
                        self.visit(inner_node)



# -*- coding: utf-8 -*-
import logging

import pytest

from .utils import make_safety_fixture, skipif_known_failing

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
_safety_fixture, _safety_state, run_cell_ = make_safety_fixture()


def run_cell(cell, cell_id=None, **kwargs):
    """Mocks the `change active cell` portion of the comm protocol"""
    if cell_id is not None:
        _safety_state[0].handle({
            'type': 'change_active_cell',
            'active_cell_id': cell_id
        })
    run_cell_(cell, **kwargs)


def test_simple():
    cells = {
        0: 'x = 0',
        1: 'y = x + 1',
        2: 'x = 42',
        3: 'logging.info(y)',
    }
    run_cell(cells[0])
    run_cell(cells[1])
    run_cell(cells[2])
    response = _safety_state[0].check_and_link_multiple_cells(cells)
    assert response['stale_cells'] == [3]
    assert response['fresh_cells'] == []
    assert response['stale_links'] == {3: [1]}
    assert response['refresher_links'] == {1: [3]}


def test_refresh_after_exception_fixed():
    cells = {
        0: 'x = 0',
        1: 'y = x + 1',
        2: 'logging.info(y)',
    }
    run_cell(cells[0], 0)
    run_cell(cells[2], 2, ignore_exceptions=True)
    run_cell(cells[1], 1)
    response = _safety_state[0].check_and_link_multiple_cells(cells)
    assert response['fresh_cells'] == [2]


def test_refresh_after_val_changed():
    cells = {
        0: 'x = 0',
        1: 'y = x + 1',
        2: 'logging.info(y)',
        3: 'y = 42',
    }
    run_cell(cells[0], 0)
    run_cell(cells[1], 1)
    run_cell(cells[2], 2)
    run_cell(cells[3], 3)
    response = _safety_state[0].check_and_link_multiple_cells(cells)
    assert response['fresh_cells'] == [2]


def test_inner_mutation_considered_fresh():
    cells = {
        0: 'lst_0 = [0,1,2]',
        1: 'lst_1 = [3,4,5]',
        2: 'lst = [lst_0, lst_1]',
        3: 'logging.info(lst)',
        4: 'lst_0.append(42)',
    }
    for idx, cell in cells.items():
        run_cell(cell, idx)
    response = _safety_state[0].check_and_link_multiple_cells(cells)
    assert response['stale_cells'] == []
    assert response['fresh_cells'] == [2, 3]


@pytest.mark.parametrize("force_subscript_symbol_creation", [True, False])
def test_update_list_elem(force_subscript_symbol_creation):
    cells = {
        0: """
class Foo(object):
    def __init__(self):
        self.counter = 0
        self.dummy = 0
        
    def inc(self):
        self.counter += 1""",

        1: """
lst = []
for i in range(5):
    x = Foo()
    lst.append(x)""",

        2: """
for foo in lst:
    foo.inc()""",

        3: 'print(lst)',
    }

    for idx, cell in cells.items():
        run_cell(cell, idx)

    response = _safety_state[0].check_and_link_multiple_cells(cells)
    assert response['stale_cells'] == []
    assert response['fresh_cells'] == []

    cells[4] = 'x.inc()'
    run_cell(cells[4], 4)

    response = _safety_state[0].check_and_link_multiple_cells(cells)
    assert response['stale_cells'] == []
    assert response['fresh_cells'] == [2, 3]

    cells[5] = 'foo.inc()'
    run_cell(cells[5], 5)
    response = _safety_state[0].check_and_link_multiple_cells(cells)
    assert response['stale_cells'] == []
    assert response['fresh_cells'] == [2, 3, 4]

    if force_subscript_symbol_creation:
        cells[6] = 'lst[-1]'
        run_cell(cells[6], 6)
        response = _safety_state[0].check_and_link_multiple_cells(cells)
        assert response['stale_cells'] == []
        assert response['fresh_cells'] == [2, 3, 4]

    run_cell(cells[4], 4)
    run_cell('%safety show_deps foo', 1234)
    response = _safety_state[0].check_and_link_multiple_cells(cells)
    assert response['stale_cells'] == []
    assert response['fresh_cells'] == [2, 3, 5] + ([6] if force_subscript_symbol_creation else [])


@skipif_known_failing
def test_no_freshness_for_alias_assignment_post_mutation():
    cells = {
        0: 'x = []',
        1: 'y = x',
        2: 'x.append(5)',
    }
    for idx, cell in cells.items():
        run_cell(cell, idx)
    response = _safety_state[0].check_and_link_multiple_cells(cells)
    assert response['stale_cells'] == []
    assert response['fresh_cells'] == []


def test_fresh_after_import():
    cells = {
        0: 'x = np.random.random(10)',
        1: 'import numpy as np'
    }
    for idx, cell in cells.items():
        run_cell(cell, idx, ignore_exceptions=True)
    response = _safety_state[0].check_and_link_multiple_cells(cells)
    assert response['stale_cells'] == []
    assert response['fresh_cells'] == [0]


def test_external_object_update_propagates_to_stale_namespace_symbols():
    cells = {
        0: 'import fakelib',
        1: 'foo = fakelib.Foo()',
        2: 'logging.info(foo.x)',
        3: 'x = 42',
        4: 'foo.x = x + 1',
        5: 'x = 43',
        6: 'foo = foo.set_x(10)',
    }
    old_skip_stale = _safety_state[0].config.get('skip_unsafe_cells', True)
    try:
        _safety_state[0].config.skip_unsafe_cells = False
        for idx, cell in cells.items():
            run_cell(cell, idx)
        response = _safety_state[0].check_and_link_multiple_cells(cells)
        assert response['stale_cells'] == []
        assert response['fresh_cells'] == [2, 4]
    finally:
        _safety_state[0].config.skip_unsafe_cells = old_skip_stale


def test_symbol_on_both_sides_of_assignment():
    cells = {
        0: 'x = 0',
        1: 'y = x + 1',
        2: 'x = 42',
    }
    for idx, cell in cells.items():
        run_cell(cell, idx)
    cells[3] = 'y += 7'
    response = _safety_state[0].check_and_link_multiple_cells(cells)
    assert response['stale_cells'] == [3]
    assert response['fresh_cells'] == [1]
    assert list(response['refresher_links'].keys()) == [1]



# -*- coding: utf-8 -*-
import logging
from typing import cast, TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Set
    from nbsafety.data_model.data_symbol import DataSymbol
    from nbsafety.data_model.scope import NamespaceScope
    from nbsafety.safety import NotebookSafety

logger = logging.getLogger(__name__)


class UpdateProtocol(object):
    def __init__(
            self,
            safety: 'NotebookSafety',
            updated_sym: 'DataSymbol',
            new_deps: 'Set[DataSymbol]',
            mutated: bool
    ):
        self.safety = safety
        self.updated_sym = updated_sym
        self.new_deps = new_deps
        self.mutated = mutated
        self.seen: Set[DataSymbol] = set()

    def __call__(self, propagate=True):
        namespace_refresh = None
        if propagate:
            if self.mutated or self.updated_sym.obj_id != self.updated_sym.cached_obj_id:
                self._collect_updated_symbols(self.updated_sym, skip_aliases=not self.mutated)
            if self.updated_sym.cached_obj_id is not None:
                # TODO: also condition on non simple assign
                namespace = self.safety.namespaces.get(self.updated_sym.obj_id, None)
                if namespace is not None:
                    # TODO: go deeper?
                    namespace_refresh = set(namespace.all_data_symbols_this_indentation())
        updated_symbols = set(self.seen)
        self.safety.updated_symbols |= updated_symbols
        self.seen |= self.new_deps  # don't propagate to stuff on RHS
        for dsym in updated_symbols:
            self._propagate_staleness_to_deps(dsym, skip_seen_check=True)
        # important! don't bump defined_cell_num until the very end!
        #  need to wait until here because, by default,
        #  we don't want to propagate to symbols defined in the same cell
        for updated_sym in updated_symbols:
            if not updated_sym.is_stale:
                updated_sym.refresh()
        self.updated_sym.refresh()
        if namespace_refresh is not None:
            for updated_sym in namespace_refresh:
                updated_sym.refresh()
                for updated_sym_alias in self.safety.aliases.get(updated_sym.obj_id, []):
                    updated_sym_alias.refresh()

    def _collect_updated_symbols(self, dsym: 'DataSymbol', skip_aliases=False):
        if dsym.is_import:
            return
        if skip_aliases:
            aliases_to_consider = {dsym}
        else:
            aliases_to_consider = self.safety.aliases[dsym.obj_id]
        for dsym_alias in aliases_to_consider:
            if dsym_alias.is_import or dsym_alias in self.seen:
                continue
            self.seen.add(dsym_alias)
            containing_scope: 'NamespaceScope' = cast('NamespaceScope', dsym_alias.containing_scope)
            if not containing_scope.is_namespace_scope:
                continue
            # TODO: figure out what this is for again
            # self.safety.updated_scopes.add(containing_scope)
            containing_scope.max_defined_timestamp = self.safety.cell_counter()
            containing_namespace_obj_id = containing_scope.obj_id
            for alias in self.safety.aliases[containing_namespace_obj_id]:
                alias.namespace_stale_symbols.discard(dsym)
                # print('discard stale', dsym, 'from', alias, 'namespace, has fresher ancestors:', alias.fresher_ancestors)
                self._collect_updated_symbols(alias)

    def _propagate_staleness_to_namespace_parents(self, dsym: 'DataSymbol', skip_seen_check=False):
        if not skip_seen_check and dsym in self.seen:
            return
        self.seen.add(dsym)
        containing_scope: 'NamespaceScope' = cast('NamespaceScope', dsym.containing_scope)
        if containing_scope is None or not containing_scope.is_namespace_scope:
            return
        for containing_alias in self.safety.aliases[containing_scope.obj_id]:
            containing_alias.namespace_stale_symbols.add(dsym)
            self._propagate_staleness_to_namespace_parents(containing_alias)
            for child in self._non_class_to_instance_children(containing_alias):
                # print('propagate from', dsym, 'to', child)
                self._propagate_staleness_to_deps(child)

    def _non_class_to_instance_children(self, dsym):
        if self.updated_sym is dsym:
            for dep_introduced_pos, dsym_children in dsym.children_by_cell_position.items():
                if not self.safety.config.get('backwards_cell_staleness_propagation', True) and dep_introduced_pos <= self.safety.active_cell_position_idx:
                    continue
                yield from dsym_children
            return
        for dep_introduced_pos, dsym_children in dsym.children_by_cell_position.items():
            if not self.safety.config.get('backwards_cell_staleness_propagation', True) and dep_introduced_pos <= self.safety.active_cell_position_idx:
                continue
            for child in dsym_children:
                # Next, complicated check to avoid propagating along a class -> instance edge.
                # The only time this is OK is when we changed the class, which will not be the case here.
                child_namespace = child.namespace
                if child_namespace is not None and child_namespace.cloned_from is not None:
                    if child_namespace.cloned_from.obj_id == dsym.obj_id:
                        continue
                yield child

    def _propagate_staleness_to_namespace_children(self, dsym: 'DataSymbol', skip_seen_check=False):
        if not skip_seen_check and dsym in self.seen:
            return
        self.seen.add(dsym)
        self_scope = self.safety.namespaces.get(dsym.obj_id, None)
        if self_scope is None:
            return
        for ns_child in self_scope.all_data_symbols_this_indentation(exclude_class=True):
            # print('propagate from', dsym, 'to namespace child', ns_child)
            self._propagate_staleness_to_deps(ns_child)

    def _propagate_staleness_to_deps(self, dsym: 'DataSymbol', skip_seen_check=False):
        if not skip_seen_check and dsym in self.seen:
            return
        self.seen.add(dsym)
        if dsym not in self.safety.updated_symbols:
            if dsym.should_mark_stale(self.updated_sym):
                dsym.fresher_ancestors.add(self.updated_sym)
                dsym.required_cell_num = self.safety.cell_counter()
                self._propagate_staleness_to_namespace_parents(dsym, skip_seen_check=True)
                self._propagate_staleness_to_namespace_children(dsym, skip_seen_check=True)
        for child in self._non_class_to_instance_children(dsym):
            # print('propagate from', dsym, 'to', child)
            self._propagate_staleness_to_deps(child)



# -*- coding: utf-8 -*-
import ast
from collections import defaultdict
import logging
from typing import Sequence, TYPE_CHECKING

from nbsafety.analysis.attr_symbols import get_attrsub_symbol_chain, AttrSubSymbolChain
from nbsafety.analysis.mixins import SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin

if TYPE_CHECKING:
    from typing import Dict, Optional, Set, Union


logger = logging.getLogger(__name__)


class TiedTuple(tuple):
    """Just a marker class indicating that we should not unpack contents of this tuple"""
    pass


def _flatten(vals):
    for v in vals:
        if isinstance(v, tuple):
            yield from _flatten(v)
        else:
            yield v


def _edges(lvals, rvals):
    if isinstance(lvals, tuple) and isinstance(rvals, tuple):
        yield from _edges_from_tuples(lvals, rvals)
    elif isinstance(lvals, tuple):
        # TODO: yield edges with subscript symbols
        for left in _flatten(lvals):
            yield left, rvals
    elif isinstance(rvals, tuple):
        # TODO: yield edges with subscript symbols
        for right in _flatten(rvals):
            yield lvals, right
    else:
        yield lvals, rvals


def _edges_from_tuples(lvals, rvals):
    if isinstance(rvals, TiedTuple):
        for lval in lvals:
            yield from _edges(lval, rvals)
    elif len(lvals) == len(rvals):
        for left, right in zip(lvals, rvals):
            yield from _edges(left, right)
    elif len(lvals) == 1:
        yield from _edges(lvals[0], rvals)
    elif len(rvals) == 1:
        yield from _edges(lvals, rvals[0])
    else:
        raise ValueError('Incompatible lists: %s, %s' % (lvals, rvals))


class GetSymbolEdges(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin, ast.NodeVisitor):
    def __init__(self):
        # TODO: figure out how to give these type annotations
        self.lval_symbols = []
        self.rval_symbols = []
        self.simple_edges = []
        self.gather_rvals = True
        self.should_overwrite = True
        self.assignment_seen = False

    def __call__(self, node: ast.AST):
        self.visit(node)
        if not self.assignment_seen:
            self._collect_simple_edges()
            self.lval_symbols = []
            self.rval_symbols = []
        yield from self.simple_edges
        for lval_list in self.lval_symbols:
            try:
                edges_for_lval = list(_edges(lval_list, tuple(self.rval_symbols)))
            except Exception as e:
                # TODO: only show warning if not in prod mode
                logger.warning('Exception occurred while computing symbol edges: %s', e)
                continue
            if len(edges_for_lval) == 0:
                for lval in lval_list:
                    yield lval, None
            else:
                yield from edges_for_lval

    def _collect_simple_edges(self):
        if len(self.lval_symbols) == 0:
            self.lval_symbols.append(None)
        if len(self.rval_symbols) == 0:
            self.rval_symbols.append(None)
        for lval in set(_flatten(self.lval_symbols)):
            for rval in set(_flatten(self.rval_symbols)):
                if lval is None and rval is None:
                    continue
                self.simple_edges.append((lval, rval))

    def gather_lvals_context(self):
        return self.push_attributes(gather_rvals=False)

    def gather_rvals_context(self):
        return self.push_attributes(gather_rvals=True)

    @property
    def to_add_set(self):
        if self.gather_rvals:
            return self.rval_symbols
        else:
            return self.lval_symbols

    @to_add_set.setter
    def to_add_set(self, val):
        if self.gather_rvals:
            self.rval_symbols = val
        else:
            self.lval_symbols = val

    def visit_Name(self, node):
        self.to_add_set.append(node.id)

    def visit_Num(self, node):
        self.visit_Constant(node)

    def visit_Str(self, node):
        self.visit_Constant(node)

    def visit_Constant(self, node):
        self.to_add_set.append(None)

    def visit_NameConstant(self, node):
        self.to_add_set.append(None)

    def visit_Tuple(self, node):
        self.visit_List_or_Tuple(node)

    def visit_List(self, node):
        self.visit_List_or_Tuple(node)

    def visit_Dict(self, node):
        temp = self.to_add_set
        self.to_add_set = []
        self.visit(node.keys)
        self.visit(node.values)
        self.to_add_set, temp = temp, self.to_add_set
        self.to_add_set.append(tuple(temp))

    def visit_List_or_Tuple(self, node):
        temp = self.to_add_set
        self.to_add_set = []
        self.visit(node.elts)
        self.to_add_set, temp = temp, self.to_add_set
        self.to_add_set.append(tuple(temp))

    def visit_expr(self, node):
        if hasattr(ast, 'NamedExpr') and isinstance(node, getattr(ast, 'NamedExpr')):
            self.visit_NamedExpr(node)
            return
        assert self.gather_rvals
        temp = self.to_add_set
        self.to_add_set = []
        super().generic_visit(node)
        self.to_add_set, temp = temp, self.to_add_set
        self.to_add_set.append(tuple(temp))

    def generic_visit(self, node: 'Union[ast.AST, Sequence[ast.AST]]'):
        # The purpose of this is to make sure we call our visit_expr method if we see an expr
        if node is None:
            return
        elif isinstance(node, ast.expr):
            self.visit_expr(node)
        else:
            super().generic_visit(node)

    def visit_Assign(self, node):
        self.assignment_seen = True
        with self.gather_lvals_context():
            for target in node.targets:
                target_lval_symbols = []
                with self.push_attributes(lval_symbols=target_lval_symbols):
                    self.visit(target)
                if isinstance(target, (ast.List, ast.Tuple)):
                    # not strictly necessary since we are robust to this later,
                    # but helps avoid unncessary double nesting, e.g., ((a, b, c),)
                    assert len(target_lval_symbols) == 1
                    assert isinstance(target_lval_symbols[0], tuple)
                    self.lval_symbols.append(target_lval_symbols[0])
                else:
                    self.lval_symbols.append(tuple(target_lval_symbols))
        with self.gather_rvals_context():
            self.visit(node.value)

    def visit_AugAssign_or_AnnAssign(self, node):
        with self.push_attributes(lval_symbols=[], rval_symbols=[]):
            with self.gather_lvals_context():
                self.visit(node.target)
            with self.gather_rvals_context():
                self.visit(node.value)
            self._collect_simple_edges()

    def visit_AnnAssign(self, node):
        self.visit_AugAssign_or_AnnAssign(node)

    def visit_AugAssign(self, node):
        self.should_overwrite = False
        self.visit_AugAssign_or_AnnAssign(node)

    def visit_Call(self, node):
        extra_to_add = []
        if isinstance(node.func, (ast.Attribute, ast.Subscript)):
            extra_to_add.append(get_attrsub_symbol_chain(node))
            to_visit = [node.args, node.keywords]
        else:
            to_visit = node
        temp = self.to_add_set
        self.to_add_set = []
        self.generic_visit(to_visit)
        self.to_add_set, temp = temp, self.to_add_set
        temp = TiedTuple(set(_flatten(temp)) | set(_flatten(extra_to_add)))
        self.to_add_set.append(temp)

    def visit_Attribute_or_Subscript(self, node):
        # TODO: we'll ignore args inside of inner calls, e.g. f.g(x, y).h
        self.to_add_set.append(get_attrsub_symbol_chain(node))

    def visit_Attribute(self, node):
        self.visit_Attribute_or_Subscript(node)

    def visit_Subscript(self, node):
        if self.gather_rvals:
            temp = self.to_add_set
            self.to_add_set = []
            self.visit_Attribute_or_Subscript(node)
            # add slice to RHS to avoid propagating to it
            self.visit(node.slice)
            self.to_add_set, temp = temp, self.to_add_set
            self.to_add_set.append(tuple(temp))
        else:
            self.visit_Attribute_or_Subscript(node)

    # def visit_Subscript(self, node):
    #     self.visit_Attribute_or_Subscript(node)
    #     # TODO: the reason we wanted this before is to avoid propagating to the slice
    #     #  add something back in to avoid propagating to everything on RHS
    #     # if self.gather_rvals:
    #     #     self.visit(node.slice)

    def visit_Keyword(self, node):
        self.visit(node.value)

    def visit_Starred(self, node):
        self.visit(node.value)

    def visit_Lambda(self, node):
        assert self.gather_rvals
        # remove node.arguments
        self.visit(node.body)
        self.visit(node.args)
        with self.push_attributes(rval_symbols=[]):
            self.visit(node.args.args)
            self.visit(node.args.vararg)
            self.visit(node.args.kwonlyargs)
            self.visit(node.args.kwarg)
            discard_set = set(self.rval_symbols)
        # throw away anything appearing in lambda body that isn't bound
        self.rval_symbols = list(set(self.rval_symbols) - discard_set)

    def visit_GeneratorExp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_DictComp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_ListComp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_SetComp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(self, node):
        assert self.gather_rvals
        to_append = set()
        for gen in node.generators:
            if isinstance(gen, ast.comprehension):
                with self.push_attributes(rval_symbols=[]):
                    self.visit(gen.iter)
                    self.visit(gen.ifs)
                    to_append |= set(_flatten(self.rval_symbols))
                with self.push_attributes(rval_symbols=[]):
                    self.visit(gen.target)
                    discard_set = set(self.rval_symbols)
            else:
                with self.push_attributes(rval_symbols=[]):
                    self.visit(gen)
                    discard_set = set(self.rval_symbols)
            to_append -= discard_set
        self.rval_symbols.append(TiedTuple(to_append))

    def visit_arg(self, node):
        self.to_add_set.append(node.arg)

    def visit_For(self, node):
        # skip body -- will have dummy since this visitor works line-by-line
        with self.push_attributes(lval_symbols=[], rval_symbols=[]):
            with self.gather_lvals_context():
                self.visit(node.target)
            with self.gather_rvals_context():
                self.visit(node.iter)
            self._collect_simple_edges()

    def visit_If(self, node):
        # skip body here too
        self.visit(node.test)

    def visit_FunctionDef_or_AsyncFunctionDef(self, node):
        with self.push_attributes(lval_symbols=[], rval_symbols=[]):
            self.lval_symbols.append(node.name)
            with self.gather_rvals_context():
                self.visit(node.args)
                self.visit(node.decorator_list)
            self._collect_simple_edges()

    def visit_FunctionDef(self, node):
        self.visit_FunctionDef_or_AsyncFunctionDef(node)

    def visit_AsyncFunctionDef(self, node):
        self.visit_FunctionDef_or_AsyncFunctionDef(node)

    def visit_ClassDef(self, node):
        with self.push_attributes(lval_symbols=[], rval_symbols=[]):
            self.lval_symbols.append(node.name)
            with self.gather_rvals_context():
                self.visit(node.bases)
                self.visit(node.decorator_list)
            self._collect_simple_edges()

    def visit_With(self, node):
        # skip body
        self.visit(node.items)

    def visit_withitem(self, node):
        with self.gather_lvals_context():
            self.visit(node.optional_vars)
        with self.gather_rvals_context():
            self.visit(node.context_expr)

    def visit_Import(self, node: 'ast.Import'):
        self.visit_Import_or_ImportFrom(node)

    def visit_ImportFrom(self, node: 'ast.ImportFrom'):
        self.visit_Import_or_ImportFrom(node)

    def visit_Import_or_ImportFrom(self, node: 'Union[ast.Import, ast.ImportFrom]'):
        with self.push_attributes(lval_symbols=[], rval_symbols=[]):
            for name in node.names:
                if name.asname is None:
                    if name.name != '*' and '.' not in name.name:
                        self.lval_symbols.append(name.name)
                else:
                    self.lval_symbols.append(name.asname)
            self._collect_simple_edges()

    def visit_NamedExpr(self, node):
        with self.push_attributes(lval_symbols=[], rval_symbols=[]):
            with self.gather_lvals_context():
                self.visit(node.target)
            with self.gather_rvals_context():
                self.visit(node.value)
            rvals_to_extend = self.lval_symbols + self.rval_symbols
            self._collect_simple_edges()
        self.rval_symbols.extend(rvals_to_extend)


def get_assignment_lval_and_rval_symbol_refs(node: 'Union[str, ast.AST]'):
    if isinstance(node, str):
        node = ast.parse(node).body[0]
    yield from GetSymbolEdges()(node)


def get_symbol_edges(node: 'Union[str, ast.AST]'):
    if isinstance(node, str):
        node = ast.parse(node).body[0]
    visitor = GetSymbolEdges()
    edges: Dict[Optional[str], Set[Optional[str]]] = defaultdict(set)
    for edge in visitor(node):
        # FIXME: figure out how to handle attributes in a principled manner here
        left, right = edge
        if isinstance(left, AttrSubSymbolChain) and isinstance(right, AttrSubSymbolChain):
            # still need this to indicate non empty edge set
            edges[None].add(None)
        elif isinstance(left, AttrSubSymbolChain):
            assert not isinstance(right, AttrSubSymbolChain)
            edges[None].add(right)
        elif isinstance(right, AttrSubSymbolChain) or right is None:
            # just get the lval in the keys
            edges[left].add(None)
            edges[left].discard(None)
        else:
            edges[left].add(right)
    return edges, visitor.should_overwrite



#!/usr/bin/env python
import argparse
import json
import os
import subprocess
import sys

from nbsafety.version import make_version_tuple


def main(args):
    components = list(make_version_tuple())
    if args.bump:
        components[-1] += 1
    version = '.'.join(str(c) for c in components)
    if args.tag:
        subprocess.check_output(['git', 'tag', version])
    for package_dot_json_loc in ['./frontend/labextension', './frontend/nbextension']:
        package_dot_json = os.path.join(package_dot_json_loc, 'package.json')
        with open(package_dot_json, 'r') as f:
            package_json = json.loads(f.read())
        if package_json.get('version', None) != version:
            package_json['version'] = version
            with open(package_dot_json, 'w') as f:
                f.write(json.dumps(package_json, indent=2))
    return 0


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Create and synchronize version tags across packages.'
    )
    parser.add_argument('--bump', action='store_true', help='Whether to increment the version.')
    parser.add_argument('--tag', action='store_true', help='Whether to increment the version.')
    args = parser.parse_args()
    sys.exit(main(args))



# -*- coding: utf-8 -*-
import ast
from contextlib import contextmanager
import sys


class FastAst(object):
    _LOCATION_OF_NODE = None

    @staticmethod
    @contextmanager
    def location_of(node):
        """
        All nodes created like `fast.AST(...)` instead of
        `ast.AST(...)` will have get location info from `node`.
        """
        old_location_of_node = FastAst._LOCATION_OF_NODE
        _LOCATION_OF_NODE = node
        yield
        _LOCATION_OF_NODE = old_location_of_node

    @staticmethod
    def kw(arg, value):
        return FastAst.keyword(arg=arg, value=value)

    @staticmethod
    def kwargs(**kwargs):
        return [FastAst.keyword(arg=arg, value=value) for arg, value in kwargs.items()]


def _make_func(func_name):
    def ctor(*args, **kwargs):
        ret = getattr(ast, func_name)(*args, **kwargs)
        if FastAst._LOCATION_OF_NODE is not None:
            ast.copy_location(ret, FastAst._LOCATION_OF_NODE)
        return ret
    return ctor


for ctor_name in ast.__dict__:
    if ctor_name.startswith('_'):
        continue
    setattr(FastAst, ctor_name, staticmethod(_make_func(ctor_name)))

if sys.version_info >= (3, 8):
    FastAst.Str = staticmethod(_make_func('Constant'))
    FastAst.Num = staticmethod(_make_func('Constant'))



import ast
import logging
import traceback
from typing import TYPE_CHECKING

from nbsafety.tracing.ast_eavesdrop import AstEavesdropper
from nbsafety.tracing.stmt_inserter import StatementInserter
from nbsafety.tracing.stmt_mapper import StatementMapper

if TYPE_CHECKING:
    from nbsafety.safety import NotebookSafety


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class SafetyAstRewriter(ast.NodeTransformer):
    def __init__(self, safety: 'NotebookSafety'):
        self.safety = safety

    def visit(self, node: 'ast.AST'):
        try:
            mapper = StatementMapper(self.safety.statement_cache[self.safety.cell_counter()], self.safety.ast_node_by_id)
            orig_to_copy_mapping = mapper(node)
            # very important that the eavesdropper does not create new ast nodes for ast.stmt (but just
            # modifies existing ones), since StatementInserter relies on being able to map these
            node = AstEavesdropper(orig_to_copy_mapping).visit(node)
            node = StatementInserter(orig_to_copy_mapping).visit(node)
        except Exception as e:
            self.safety.set_ast_transformer_raised(e)
            traceback.print_exc()
            raise e
        return node



# -*- coding: utf-8 -*-
from nbsafety._version import get_versions
__version__ = get_versions()['version']
del get_versions


def make_version_tuple(vstr=None):
    if vstr is None:
        vstr = __version__
    if vstr[0] == 'v':
        vstr = vstr[1:]
    components = []
    for component in vstr.split('+')[0].split('.'):
        try:
            components.append(int(component))
        except ValueError:
            break
    return tuple(components)



# -*- coding: utf-8 -*-

# Jupyter Extension points
def _jupyter_nbextension_paths():
    return [dict(
        section="notebook",
        # the path is relative to the `my_fancy_module` directory
        src="resources/nbextension",
        # directory in the `nbextension/` namespace
        dest="nbsafety",
        # _also_ in the `nbextension/` namespace
        require="nbsafety/index")]


def load_jupyter_server_extension(nbapp):
    nbapp.log.info("my module enabled!")



# -*- coding: utf-8 -*-
import logging

from test.utils import make_safety_fixture

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
_safety_fixture, _safety_state, run_cell_ = make_safety_fixture()


def updated_symbol_names():
    return sorted(map(lambda sym: sym.readable_name, _safety_state[0].updated_symbols))


def run_cell(cell):
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell)


def test_simplest():
    run_cell('a = 0')
    assert updated_symbol_names() == ['a']
    run_cell('a += 1')
    assert updated_symbol_names() == ['a']


def test_dict_hierarchy():
    run_cell('d = {}')
    assert updated_symbol_names() == ['d']
    run_cell('d["foo"] = {}')
    assert updated_symbol_names() == sorted(['d[foo]', 'd'])
    run_cell('d["foo"]["bar"] = []')
    assert updated_symbol_names() == sorted(['d[foo][bar]', 'd[foo]', 'd'])
    run_cell('d["foo"]["bar"] = 0')
    assert updated_symbol_names() == sorted(['d[foo][bar]', 'd[foo]', 'd'])



# -*- coding: utf-8 -*-
import ast
from collections import defaultdict
from contextlib import contextmanager
import inspect
import logging
import re
from typing import cast, TYPE_CHECKING

from IPython import get_ipython
from IPython.core.magic import register_cell_magic, register_line_magic

from nbsafety.analysis import (
    compute_live_dead_symbol_refs,
    compute_call_chain_live_symbols,
    get_symbols_for_references,
)
from nbsafety.ipython_utils import (
    ast_transformer_context,
    cell_counter,
    run_cell,
    save_number_of_currently_executing_cell,
)
from nbsafety import line_magics
from nbsafety.data_model.scope import Scope, NamespaceScope
from nbsafety.run_mode import SafetyRunMode
from nbsafety.tracing import SafetyAstRewriter, TracingManager
from nbsafety.utils import DotDict

if TYPE_CHECKING:
    from typing import Any, Dict, List, Set, Optional, Tuple, Union
    from types import FrameType
    from nbsafety.data_model.data_symbol import DataSymbol
    CellId = Union[str, int]

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)

_MAX_WARNINGS = 10
_SAFETY_LINE_MAGIC = 'safety'

_NB_MAGIC_PATTERN = re.compile(r'(^%|^!|^cd |\?$)')


def _safety_warning(node: 'DataSymbol'):
    if not node.is_stale:
        raise ValueError('Expected node with stale ancestor; got %s' % node)
    if node.defined_cell_num < 1:
        return
    fresher_symbols = node.fresher_ancestors
    if len(fresher_symbols) == 0:
        fresher_symbols = node.namespace_stale_symbols
    logger.warning(
        f'`{node.readable_name}` defined in cell {node.defined_cell_num} may depend on '
        f'old version(s) of [{", ".join(f"`{str(dep)}`" for dep in fresher_symbols)}] '
        f'(latest update in cell {node.required_cell_num}).'
        f'\n\n(Run cell again to override and execute anyway.)'
    )


class NotebookSafety(object):
    """Holds all the state necessary to detect stale dependencies in Jupyter notebooks."""
    def __init__(self, cell_magic_name=None, use_comm=False, **kwargs):
        self.config = DotDict(dict(
            store_history=kwargs.pop('store_history', True),
            test_context=kwargs.pop('test_context', False),
            use_comm=use_comm,
            trace_messages_enabled=kwargs.pop('trace_messages_enabled', False),
            backwards_cell_staleness_propagation=True,
            track_dependencies=True,
            naive_refresher_computation=False,
            skip_unsafe_cells=kwargs.pop('skip_unsafe', True),
            mode=SafetyRunMode.get(),
            **kwargs
        ))
        # Note: explicitly adding the types helps PyCharm's built-in code inspection
        self.namespaces: 'Dict[int, NamespaceScope]' = {}
        self.aliases: 'Dict[int, Set[DataSymbol]]' = defaultdict(set)
        self.global_scope: 'Scope' = Scope(self)
        self.updated_symbols: 'Set[DataSymbol]' = set()
        self.updated_scopes: 'Set[NamespaceScope]' = set()
        self.garbage_namespace_obj_ids: 'Set[int]' = set()
        self.ast_node_by_id: 'Dict[int, ast.AST]' = {}
        self.statement_cache: 'Dict[int, Dict[int, ast.stmt]]' = defaultdict(dict)
        self.statement_to_func_cell: 'Dict[int, DataSymbol]' = {}
        self.tracing_manager: 'TracingManager' = TracingManager(self)
        self.stale_dependency_detected = False
        self.active_cell_position_idx = -1
        self._last_execution_counter = 0
        self._counters_by_cell_id: Dict[CellId, int] = {}
        self._active_cell_id: Optional[str] = None
        if cell_magic_name is None:
            self._cell_magic = None
        else:
            self._cell_magic = self._make_cell_magic(cell_magic_name)
        self._line_magic = self._make_line_magic()
        self._last_refused_code: Optional[str] = None
        self._prev_cell_stale_symbols: Set[DataSymbol] = set()
        self._cell_counter = 1
        self._recorded_cell_name_to_cell_num = True
        self._cell_name_to_cell_num_mapping: 'Dict[str, int]' = {}
        self._ast_transformer_raised: 'Optional[Exception]' = None
        if use_comm:
            get_ipython().kernel.comm_manager.register_target(__package__, self._comm_target)

    @property
    def is_develop(self) -> bool:
        return self.config.get('mode', SafetyRunMode.PRODUCTION) == SafetyRunMode.DEVELOP

    @property
    def is_test(self) -> bool:
        return self.config.get('test_context', False)

    def cell_counter(self):
        if self.config.store_history:
            return cell_counter()
        else:
            return self._cell_counter

    def set_ast_transformer_raised(self, new_val: 'Optional[Exception]' = None) -> 'Optional[Exception]':
        ret = self._ast_transformer_raised
        self._ast_transformer_raised = new_val
        return ret

    def get_position(self, frame: 'FrameType'):
        cell_num = self._cell_name_to_cell_num_mapping[frame.f_code.co_filename.split('-')[3]]
        return cell_num, frame.f_lineno

    def maybe_set_name_to_cell_num_mapping(self, frame: 'FrameType'):
        if self._recorded_cell_name_to_cell_num:
            return
        self._recorded_cell_name_to_cell_num = True
        self._cell_name_to_cell_num_mapping[frame.f_code.co_filename.split('-')[3]] = self.cell_counter()

    def set_active_cell(self, cell_id, position_idx=-1):
        self._active_cell_id = cell_id
        self.active_cell_position_idx = position_idx

    def _comm_target(self, comm, open_msg):
        @comm.on_msg
        def _responder(msg):
            request = msg['content']['data']
            self.handle(request, comm=comm)

        comm.send({'type': 'establish'})

    def handle(self, request, comm=None):
        if request['type'] == 'change_active_cell':
            self.set_active_cell(request['active_cell_id'], position_idx=request.get('active_cell_order_idx', -1))
        elif request['type'] == 'cell_freshness':
            cell_id = request.get('executed_cell_id', None)
            if cell_id is not None:
                self._counters_by_cell_id[cell_id] = self._last_execution_counter
            cells_by_id = request['content_by_cell_id']
            if self.config.get('backwards_cell_staleness_propagation', True):
                order_index_by_id = None
                last_cell_exec_position_idx = -1
            else:
                order_index_by_id = request['order_index_by_cell_id']
                last_cell_exec_position_idx = order_index_by_id.get(cell_id, -1)
            response = self.check_and_link_multiple_cells(cells_by_id, order_index_by_id)
            response['type'] = 'cell_freshness'
            response['last_cell_exec_position_idx'] = last_cell_exec_position_idx
            if comm is not None:
                comm.send(response)
        else:
            logger.error('Unsupported request type for request %s' % request)

    def check_and_link_multiple_cells(
            self,
            cells_by_id: 'Dict[CellId, str]',
            order_index_by_cell_id: 'Optional[Dict[CellId, int]]' = None
    ) -> 'Dict[str, Any]':
        stale_cells = set()
        fresh_cells = []
        stale_symbols_by_cell_id: 'Dict[CellId, Set[DataSymbol]]' = {}
        killing_cell_ids_for_symbol: 'Dict[DataSymbol, Set[CellId]]' = defaultdict(set)
        for cell_id, cell_content in cells_by_id.items():
            if (order_index_by_cell_id is not None and
                    order_index_by_cell_id.get(cell_id, -1) <= self.active_cell_position_idx):
                continue
            try:
                symbols = self._check_cell_and_resolve_symbols(cell_content)
                stale_symbols, dead_symbols = symbols['stale'], symbols['dead']
                if len(stale_symbols) > 0:
                    stale_symbols_by_cell_id[cell_id] = stale_symbols
                    stale_cells.add(cell_id)
                elif (self._get_max_defined_cell_num_for_symbols(symbols['live']) >
                      self._counters_by_cell_id.get(cell_id, cast(int, float('inf')))):
                    fresh_cells.append(cell_id)
                for dead_sym in dead_symbols:
                    killing_cell_ids_for_symbol[dead_sym].add(cell_id)
            except SyntaxError:
                continue
        stale_links: 'Dict[CellId, Set[CellId]]' = defaultdict(set)
        refresher_links: 'Dict[CellId, List[CellId]]' = defaultdict(list)
        for stale_cell_id in stale_cells:
            stale_syms = stale_symbols_by_cell_id[stale_cell_id]
            if self.config.get('naive_refresher_computation', False):
                refresher_cell_ids = self._naive_compute_refresher_cells(
                    stale_cell_id,
                    stale_syms,
                    cells_by_id,
                    order_index_by_cell_id=order_index_by_cell_id
                )
            else:
                refresher_cell_ids = set.union(*(killing_cell_ids_for_symbol[stale_sym] for stale_sym in stale_syms))
            stale_links[stale_cell_id] = refresher_cell_ids
        stale_link_changes = True
        # transitive closer up until we hit non-stale refresher cells
        while stale_link_changes:
            stale_link_changes = False
            for stale_cell_id in stale_cells:
                new_stale_links = set(stale_links[stale_cell_id])
                original_length = len(new_stale_links)
                for refresher_cell_id in stale_links[stale_cell_id]:
                    if refresher_cell_id not in stale_cells:
                        continue
                    new_stale_links |= stale_links[refresher_cell_id]
                new_stale_links.discard(stale_cell_id)
                stale_link_changes = stale_link_changes or original_length != len(new_stale_links)
                stale_links[stale_cell_id] = new_stale_links
        for stale_cell_id in stale_cells:
            stale_links[stale_cell_id] -= stale_cells
            for refresher_cell_id in stale_links[stale_cell_id]:
                refresher_links[refresher_cell_id].append(stale_cell_id)
        return {
            'stale_cells': list(stale_cells),
            'fresh_cells': fresh_cells,
            'stale_links': {
                stale_cell_id: list(refresher_cell_ids)
                for stale_cell_id, refresher_cell_ids in stale_links.items()
            },
            'refresher_links': refresher_links,
        }

    def _naive_compute_refresher_cells(
            self,
            stale_cell_id: 'CellId',
            stale_symbols: 'Set[DataSymbol]',
            cells_by_id: 'Dict[CellId, str]',
            order_index_by_cell_id: 'Optional[Dict[CellId, int]]' = None
    ) -> 'Set[CellId]':
        refresher_cell_ids: Set[CellId] = set()
        stale_cell_content = cells_by_id[stale_cell_id]
        for cell_id, cell_content in cells_by_id.items():
            if cell_id == stale_cell_id:
                continue
            if (order_index_by_cell_id is not None and
                    order_index_by_cell_id.get(cell_id, -1) >= order_index_by_cell_id.get(stale_cell_id, -1)):
                continue
            concated_content = f'{cell_content}\n\n{stale_cell_content}'
            try:
                concated_stale_symbols = self._check_cell_and_resolve_symbols(concated_content)['stale']
            except SyntaxError:
                continue
            if concated_stale_symbols < stale_symbols:
                refresher_cell_ids.add(cell_id)
        return refresher_cell_ids

    @staticmethod
    def _get_cell_ast(cell):
        lines = []
        for line in cell.strip().split('\n'):
            # TODO: figure out more robust strategy for filtering / transforming lines for the ast parser
            # we filter line magics, but for %time, we would ideally like to trace the statement being timed
            # TODO: how to do this?
            if _NB_MAGIC_PATTERN.search(line) is None:
                lines.append(line)
        return ast.parse('\n'.join(lines))

    def _get_max_defined_cell_num_for_symbols(self, symbols: 'Set[DataSymbol]') -> int:
        max_defined_cell_num = -1
        for dsym in symbols:
            max_defined_cell_num = max(max_defined_cell_num, dsym.defined_cell_num)
            if dsym.obj_id in self.namespaces:
                namespace_scope = self.namespaces[dsym.obj_id]
                max_defined_cell_num = max(max_defined_cell_num, namespace_scope.max_defined_timestamp)
        return max_defined_cell_num

    def _check_cell_and_resolve_symbols(
            self,
            cell: 'Union[ast.Module, str]'
    ) -> 'Dict[str, Set[DataSymbol]]':
        if isinstance(cell, str):
            cell = self._get_cell_ast(cell)
        live_symbol_refs, dead_symbol_refs = compute_live_dead_symbol_refs(cell)
        live_symbols, called_symbols = get_symbols_for_references(live_symbol_refs, self.global_scope)
        live_symbols = live_symbols.union(compute_call_chain_live_symbols(called_symbols))
        # only mark dead attrsubs as killed if we can traverse the entire chain
        dead_symbols, _ = get_symbols_for_references(
            dead_symbol_refs, self.global_scope, only_add_successful_resolutions=True
        )
        stale_symbols = set(dsym for dsym in live_symbols if dsym.is_stale)
        return {
            'live': live_symbols,
            'dead': dead_symbols,
            'stale': stale_symbols,
        }

    def _precheck_for_stale(self, cell: str):
        # Precheck process. First obtain the names that need to be checked. Then we check if their
        # `defined_cell_num` is greater than or equal to required; if not we give a warning and return `True`.
        try:
            cell_ast = self._get_cell_ast(cell)
        except SyntaxError:
            return False
        symbols = self._check_cell_and_resolve_symbols(cell_ast)
        stale_symbols, live_symbols = symbols['stale'], symbols['live']
        if self._last_refused_code is None or cell != self._last_refused_code:
            self._prev_cell_stale_symbols = stale_symbols
            if len(stale_symbols) > 0:
                warning_counter = 0
                for node in self._prev_cell_stale_symbols:
                    if warning_counter >= _MAX_WARNINGS:
                        logger.warning(f'{len(self._prev_cell_stale_symbols) - warning_counter}'
                                       ' more nodes with stale dependencies skipped...')
                        break
                    _safety_warning(node)
                    warning_counter += 1
                self.stale_dependency_detected = True
                self._last_refused_code = cell
                return True
        else:
            # Instead of breaking the dependency chain, simply refresh the nodes
            # with stale deps to their required cell numbers
            for node in self._prev_cell_stale_symbols:
                node.defined_cell_num = node.required_cell_num
                node.namespace_stale_symbols = set()
                node.fresher_ancestors = set()
            self._prev_cell_stale_symbols.clear()

        self._last_refused_code = None
        return False

    def _resync_symbols(self, symbols: 'Set[DataSymbol]'):
        for dsym in symbols:
            if not dsym.containing_scope.is_global:
                continue
            obj = get_ipython().user_global_ns.get(dsym.name, None)
            if obj is None:
                continue
            if dsym.obj_id == id(obj):
                continue
            for alias in self.aliases[dsym.cached_obj_id] | self.aliases[dsym.obj_id]:
                if not alias.containing_scope.is_namespace_scope:
                    continue
                containing_scope = cast(NamespaceScope, alias.containing_scope)
                if containing_scope._obj_ref is None or containing_scope._obj_ref() is None:
                    continue
                containing_obj = containing_scope._obj_ref()
                # TODO: handle dict case too
                if isinstance(containing_obj, list) and containing_obj[-1] is obj:
                    # new_alias_sym = containing_scope.upsert_data_symbol_for_name(
                    containing_scope.upsert_data_symbol_for_name(
                        len(containing_obj) - 1,
                        obj,
                        set(alias.parents),
                        alias.stmt_node,
                        is_subscript=True,
                        propagate=False
                    )
                    # self.aliases[id(obj)].add(new_alias_sym)
            self.aliases[dsym.cached_obj_id].discard(dsym)
            self.aliases[dsym.obj_id].discard(dsym)
            self.aliases[id(obj)].add(dsym)
            namespace = self.namespaces.get(dsym.obj_id, None)
            if namespace is not None:
                namespace.update_obj_ref(obj)
                del self.namespaces[dsym.obj_id]
                self.namespaces[id(obj)] = namespace
            dsym.update_obj_ref(obj)

    def safe_execute(self, cell: str, run_cell_func):
        with save_number_of_currently_executing_cell():
            self._last_execution_counter = self.cell_counter()

            if self._active_cell_id is not None:
                self._counters_by_cell_id[self._active_cell_id] = self._last_execution_counter
                self._active_cell_id = None
            # Stage 1: Precheck.
            if self._precheck_for_stale(cell) and self.config.get('skip_unsafe_cells', True):
                # FIXME: hack to increase cell number
                #  ideally we shouldn't show a cell number at all if we fail precheck since nothing executed
                return run_cell_func('None')

            # Stage 2: Trace / run the cell, updating dependencies as they are encountered.
            try:
                with self._tracing_context():
                    ret = run_cell_func(cell)
                # Stage 2.1: resync any defined symbols that could have gotten out-of-sync
                #  due to tracing being disabled
                defined = self._check_cell_and_resolve_symbols(cell)['dead']
                self._resync_symbols(defined)
            finally:
                if not self.config.store_history:
                    self._cell_counter += 1
                return ret

    def _make_cell_magic(self, cell_magic_name):
        def _run_cell_func(cell):
            run_cell(cell, store_history=self.config.store_history)

        def _dependency_safety(_, cell: str):
            self.safe_execute(cell, _run_cell_func)

        # FIXME (smacke): probably not a great idea to rely on this
        _dependency_safety.__name__ = cell_magic_name
        return register_cell_magic(_dependency_safety)

    @contextmanager
    def _tracing_context(self):
        self.updated_symbols.clear()
        self.updated_scopes.clear()
        self._recorded_cell_name_to_cell_num = False

        try:
            with self.tracing_manager.tracing_context():
                with ast_transformer_context([SafetyAstRewriter(self)]):
                    yield
        finally:
            # TODO: actually handle errors that occurred in our code while tracing
            # if not self.trace_state_manager.error_occurred:
            self._reset_trace_state_hook()

    def _reset_trace_state_hook(self):
        # this assert doesn't hold anymore now that tracing could be disabled inside of something
        # assert len(self.attr_trace_manager.stack) == 0
        self.tracing_manager = TracingManager(self)
        self._gc()

    def _make_line_magic(self):
        line_magic_names = [f[0] for f in inspect.getmembers(line_magics) if inspect.isfunction(f[1])]

        def _safety(line_: str):
            line = line_.split()
            if not line or line[0] not in line_magic_names:
                print(line_magics.USAGE)
                return
            elif line[0] in ("show_deps", "show_dependency", "show_dependencies"):
                return line_magics.show_deps(self, line)
            elif line[0] == "show_stale":
                return line_magics.show_stale(self, line)
            elif line[0] == "trace_messages":
                return line_magics.trace_messages(self, line)
            elif line[0] == "remove_dependency":
                return line_magics.remove_dep(self, line)
            elif line[0] in ("add_dependency", "add_dep"):
                return line_magics.add_dep(self, line)
            elif line[0] == "turn_off_warnings_for":
                return line_magics.turn_off_warnings_for(self, line)
            elif line[0] == "turn_on_warnings_for":
                return line_magics.turn_on_warnings_for(self, line)

        # FIXME (smacke): probably not a great idea to rely on this
        _safety.__name__ = _SAFETY_LINE_MAGIC
        return register_line_magic(_safety)

    @property
    def dependency_tracking_enabled(self):
        return self.config.get('track_dependencies', True)

    @property
    def cell_magic_name(self):
        return self._cell_magic.__name__

    @property
    def line_magic_name(self):
        return self._line_magic.__name__

    def all_data_symbols(self):
        for alias_set in self.aliases.values():
            for alias in alias_set:
                yield alias

    def test_and_clear_detected_flag(self):
        ret = self.stale_dependency_detected
        self.stale_dependency_detected = False
        return ret

    def _namespace_gc(self):
        for obj_id in self.garbage_namespace_obj_ids:
            garbage_ns = self.namespaces.pop(obj_id, None)
            if garbage_ns is not None:
                garbage_ns.clear_namespace(obj_id)
        self.garbage_namespace_obj_ids.clear()
        # while True:
        #     for obj_id in self.garbage_namespace_obj_ids:
        #         self.namespaces.pop(obj_id, None)
        #     self.garbage_namespace_obj_ids.clear()
        #     for obj_id, namespace in self.namespaces.items():
        #         if namespace.is_garbage:
        #             self.garbage_namespace_obj_ids.add(namespace.obj_id)
        #     if len(self.garbage_namespace_obj_ids) == 0:
        #         break

    def _gc(self):
        for dsym in list(self.all_data_symbols()):
            if dsym.is_garbage:
                dsym.collect_self_garbage()

    def retrieve_namespace_attr_or_sub(self, obj: 'Any', attr_or_sub: 'Union[str, int]', is_subscript: bool):
        try:
            if is_subscript:
                # TODO: more complete list of things that are checkable
                #  or could cause side effects upon subscripting
                return obj[attr_or_sub]
            else:
                if self.is_develop:
                    assert isinstance(attr_or_sub, str)
                return getattr(obj, cast(str, attr_or_sub))
        except (AttributeError, IndexError, KeyError):
            raise
        except Exception as e:
            if self.is_develop:
                logger.warning('unexpected exception: %s', e)
                logger.warning('object: %s', obj)
                logger.warning('attr / subscript: %s', attr_or_sub)
            raise e



# -*- coding: utf-8 -*-
import ast
import logging
import sys
from typing import cast, TYPE_CHECKING

from nbsafety.analysis.attr_symbols import AttrSubSymbolChain
from nbsafety.analysis.live_refs import compute_live_dead_symbol_refs

if TYPE_CHECKING:
    from nbsafety.data_model.data_symbol import DataSymbol
    from nbsafety.data_model.scope import Scope
    from nbsafety.types import SymbolRef
    from typing import Set, Tuple

logger = logging.getLogger(__name__)


def get_symbols_for_references(
        symbol_refs: 'Set[SymbolRef]',
        scope: 'Scope',
        only_add_successful_resolutions: bool = False,
) -> 'Tuple[Set[DataSymbol], Set[DataSymbol]]':
    symbols = set()
    called_symbols = set()
    for symbol_ref in symbol_refs:
        success = True
        if isinstance(symbol_ref, str):
            dsym = scope.lookup_data_symbol_by_name(symbol_ref)
            called_dsym = None
        elif isinstance(symbol_ref, AttrSubSymbolChain):
            dsym, called_dsym, success = scope.get_most_specific_data_symbol_for_attrsub_chain(symbol_ref)
        else:
            logger.warning('invalid type for ref %s', symbol_ref)
            continue
        if dsym is not None:
            if success or not only_add_successful_resolutions:
                symbols.add(dsym)
        if called_dsym is not None:
            called_symbols.add(called_dsym)
    return symbols, called_symbols


def compute_call_chain_live_symbols(live: 'Set[DataSymbol]'):
    seen = set()
    worklist = list(live)
    while len(worklist) > 0:
        called_dsym = worklist.pop()
        if called_dsym in seen:
            continue
        seen.add(called_dsym)
        # TODO: handle callable classes
        if not called_dsym.is_function:
            continue
        live_refs, _ = compute_live_dead_symbol_refs(
            cast(ast.FunctionDef, called_dsym.stmt_node).body, called_dsym.get_call_args()
        )
        live_symbols, called_symbols = get_symbols_for_references(live_refs, called_dsym.call_scope)
        worklist.extend(called_symbols)
        live_symbols = set(sym for sym in live_symbols if sym.is_globally_accessible)
        called_symbols = set(sym for sym in called_symbols if sym.is_globally_accessible)
        live |= live_symbols.union(called_symbols)
    return live


class ContainsNamedExprVisitor(ast.NodeVisitor):
    def __init__(self):
        self.contains_named_expr = False

    def __call__(self, node: 'ast.stmt') -> bool:
        if sys.version_info.minor < 8:
            return False
        self.visit(node)
        return self.contains_named_expr

    def visit_NamedExpr(self, node):
        self.contains_named_expr = True

    def generic_visit(self, node: 'ast.AST'):
        if self.contains_named_expr:
            return
        super().generic_visit(node)


def stmt_contains_lval(node: 'ast.stmt'):
    # TODO: expand to method calls, etc.
    simple_contains_lval = isinstance(node, (
        ast.Assign,
        ast.AnnAssign,
        ast.AugAssign,
        ast.ClassDef,
        ast.FunctionDef,
        ast.AsyncFunctionDef,
        ast.For,
        ast.Import,
        ast.ImportFrom,
        ast.With,
    ))
    return simple_contains_lval or ContainsNamedExprVisitor()(node)



# -*- coding: utf-8 -*-
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional
    import ast

class FastAst:
    _LOCATION_OF_NODE: 'Optional[ast.AST]' = None

    @staticmethod
    def location_of(*args, **kwargs) -> 'Any': ...

    @staticmethod
    def kw(arg: str, value: 'ast.expr') -> 'ast.keyword': ...

    @staticmethod
    def kwargs(**kwargs: 'ast.expr') -> 'List[ast.keyword]': ...

    @staticmethod
    def parse(*args, **kwargs) -> 'ast.Module': ...

    @staticmethod
    def Call(*args, **kwargs) -> 'ast.Call': ...
    @staticmethod
    def Name(*args, **kwargs) -> 'ast.Name': ...
    @staticmethod
    def NameConstant(*args, **kwargs) -> 'ast.NameConstant': ...
    @staticmethod
    def Subscript(*args, **kwargs) -> 'ast.Subscript': ...
    @staticmethod
    def Tuple(*args, **kwargs) -> 'ast.Tuple': ...
    @staticmethod
    def keyword(*args, **kwargs) -> 'ast.keyword': ...

    if sys.version_info < (3, 9):
        @staticmethod
        def Index(*args, **kwargs) -> 'ast.Index': ...

    if sys.version_info < (3, 8):
        @staticmethod
        def Num(*args, **kwargs) -> 'ast.Num': ...
        @staticmethod
        def Str(*args, **kwargs) -> 'ast.Str': ...
    else:
        @staticmethod
        def Num(*args, **kwargs) -> 'ast.Constant': ...
        @staticmethod
        def Str(*args, **kwargs) -> 'ast.Constant': ...
        @staticmethod
        def Constant(*args, **kwargs) -> 'ast.Constant': ...


# -*- coding: utf-8 -*-
import ast
from contextlib import contextmanager
import logging
from typing import cast, TYPE_CHECKING
import sys

from nbsafety.tracing.trace_events import TraceEvent, EMIT_EVENT
from nbsafety.utils import fast

if TYPE_CHECKING:
    from typing import Dict, List, Set, Union


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class AstEavesdropper(ast.NodeTransformer):
    def __init__(self, orig_to_copy_mapping: 'Dict[int, ast.AST]'):
        self._orig_to_copy_mapping = orig_to_copy_mapping
        self._inside_attrsub_load_chain = False

    def _emitter_ast(self):
        return fast.Name(EMIT_EVENT, ast.Load())

    def _get_copy_id_ast(self, orig_node_id: 'Union[int, ast.AST]'):
        if not isinstance(orig_node_id, int):
            orig_node_id = id(orig_node_id)
        return fast.Num(id(self._orig_to_copy_mapping[orig_node_id]))

    def visit(self, node: 'ast.AST'):
        ret = super().visit(node)
        if isinstance(node, ast.stmt):
            # we haven't inserted statements yet, and StatementInserter needs the previous ids to be identical
            assert ret is node
        return ret

    @contextmanager
    def attrsub_load_context(self, override=True):
        old = self._inside_attrsub_load_chain
        self._inside_attrsub_load_chain = override
        yield
        self._inside_attrsub_load_chain = old

    def visit_Attribute(self, node: 'ast.Attribute', call_context=False):
        return self.visit_Attribute_or_Subscript(node, call_context)

    def visit_Subscript(self, node: 'ast.Subscript', call_context=False):
        return self.visit_Attribute_or_Subscript(node, call_context)

    def visit_Attribute_or_Subscript(self, node: 'Union[ast.Attribute, ast.Subscript]', call_context=False):
        with fast.location_of(node.value):
            is_load = isinstance(node.ctx, ast.Load)
            is_subscript = isinstance(node, ast.Subscript)
            # TODO: expand beyond simple slices
            if is_subscript:
                sub_node = cast(ast.Subscript, node)
                if isinstance(sub_node.slice, ast.Index):
                    attr_or_sub = sub_node.slice.value  # type: ignore
                    # ast.copy_location(attr_or_sub, sub_node.slice.value)
                    # if isinstance(attr_or_sub, ast.Str):
                    #     attr_or_sub = attr_or_sub.s
                    # elif isinstance(attr_or_sub, ast.Num):
                    #     attr_or_sub = attr_or_sub.n
                    # else:
                    #     logger.debug('unimpled index: %s', attr_or_sub)
                    #     return node
                elif isinstance(sub_node.slice, ast.Constant):
                    # Python > 3.8 doesn't use ast.Index for constant slices
                    attr_or_sub = sub_node.slice
                else:
                    logger.debug('unimpled slice: %s', sub_node.slice)
                    return node
                # elif isinstance(sub_node.slice, ast.Slice):
                #     raise ValueError('unimpled slice: %s' % sub_node.slice)
                # elif isinstance(sub_node.slice, ast.ExtSlice):
                #     raise ValueError('unimpled slice: %s' % sub_node.slice)
                # else:
                #     raise ValueError('unexpected slice: %s' % sub_node.slice)
            else:
                attr_node = cast(ast.Attribute, node)
                attr_or_sub = fast.Str(attr_node.attr)

            extra_args: 'List[ast.keyword]' = []
            if isinstance(node.value, ast.Name):
                extra_args = fast.kwargs(name=fast.Str(node.value.id))

            with self.attrsub_load_context():
                node.value = fast.Call(
                    func=self._emitter_ast(),
                    args=[
                        TraceEvent.subscript.to_ast() if is_subscript else TraceEvent.attribute.to_ast(),
                        self._get_copy_id_ast(node.value)
                    ],
                    keywords=fast.kwargs(
                        obj=self.visit(node.value),
                        attr_or_sub=attr_or_sub,
                        ctx=fast.Str(node.ctx.__class__.__name__),
                        call_context=fast.NameConstant(call_context),
                    ) + extra_args
                )
        # end fast.location_of(node.value)
        if not self._inside_attrsub_load_chain and is_load:
            with fast.location_of(node):
                return fast.Call(
                    func=self._emitter_ast(),
                    args=[TraceEvent.after_attrsub_chain.to_ast(), self._get_copy_id_ast(node)],
                    keywords=fast.kwargs(obj=node, call_context=fast.NameConstant(call_context)),
                )
        return node

    def _get_replacement_args(self, args, should_record: bool, keywords: bool):
        replacement_args = []
        for arg in args:
            if keywords:
                maybe_kwarg = getattr(arg, 'value')
            else:
                maybe_kwarg = arg
            with fast.location_of(maybe_kwarg):
                with self.attrsub_load_context(False):
                    visited_maybe_kwarg = self.visit(maybe_kwarg)
                if should_record:
                    with self.attrsub_load_context(False):
                        new_arg_value = cast(ast.expr, fast.Call(
                            func=self._emitter_ast(),
                            args=[TraceEvent.argument.to_ast(), self._get_copy_id_ast(maybe_kwarg)],
                            keywords=fast.kwargs(obj=visited_maybe_kwarg),
                        ))
                else:
                    new_arg_value = visited_maybe_kwarg
            if keywords:
                setattr(arg, 'value', new_arg_value)
            else:
                arg = new_arg_value
            replacement_args.append(arg)
        return replacement_args

    def visit_Call(self, node: ast.Call):
        orig_node_id = id(node)
        is_attrsub = False
        if isinstance(node.func, (ast.Attribute, ast.Subscript)):
            is_attrsub = True
            with self.attrsub_load_context():
                node.func = self.visit_Attribute_or_Subscript(node.func, call_context=True)

            # TODO: need a way to rewrite ast of attribute and subscript args,
            #  and to process these separately from outer rewrite

        node.args = self._get_replacement_args(node.args, is_attrsub, False)
        node.keywords = self._get_replacement_args(node.keywords, is_attrsub, True)

        # in order to ensure that the args are processed with appropriate active scope,
        # we need to make sure not to use the active namespace scope on args (in the case
        # of a function call on an ast.Attribute).
        #
        # We do so by emitting an "enter argument list", whose handler pushes the current active
        # scope while we process each argument. The "end argument list" event will then restore
        # the active scope.
        #
        # This effectively rewrites function calls as follows:
        # f(a, b, ..., c) -> trace(f, 'enter argument list')(a, b, ..., c)
        with fast.location_of(node.func):
            node.func = fast.Call(
                func=self._emitter_ast(),
                args=[TraceEvent.before_arg_list.to_ast(), self._get_copy_id_ast(node.func)],
                keywords=fast.kwargs(obj=node.func),
            )

        # f(a, b, ..., c) -> trace(f(a, b, ..., c), 'exit argument list')
        with fast.location_of(node):
            node = fast.Call(
                func=self._emitter_ast(),
                args=[TraceEvent.after_arg_list.to_ast(), self._get_copy_id_ast(node)],
                keywords=fast.kwargs(
                    obj=node,
                    is_attrsub=fast.NameConstant(is_attrsub),
                    inside_chain=fast.NameConstant(self._inside_attrsub_load_chain),
                ),
            )

        if self._inside_attrsub_load_chain or not is_attrsub:
            return node

        with fast.location_of(node):
            return fast.Call(
                func=self._emitter_ast(),
                args=[TraceEvent.after_attrsub_chain.to_ast(), self._get_copy_id_ast(orig_node_id)],
                keywords=fast.kwargs(obj=node, call_context=fast.NameConstant(True)),
            )

    def visit_Assign(self, node: ast.Assign):
        if not isinstance(node.value, (ast.List, ast.Tuple)):
            return self.generic_visit(node)

        orig_node_value = id(node.value)
        new_targets = []
        for target in node.targets:
            new_targets.append(self.visit(target))
        node.targets = cast('List[ast.expr]', new_targets)
        with fast.location_of(node.value):
            # TODO: replace 42 with start literal tracer
            node.value = fast.Tuple([fast.Call(
                func=self._emitter_ast(),
                args=[TraceEvent.before_literal.to_ast(), self._get_copy_id_ast(orig_node_value)],
                keywords=[],
            ), node.value], ast.Load())
            slc: 'Union[ast.Constant, ast.Num, ast.Index]' = fast.Num(1)
            if sys.version_info < (3, 9):
                slc = fast.Index(slc)
            node.value = fast.Subscript(node.value, slc, ast.Load())
            node.value = fast.Call(
                func=self._emitter_ast(),
                args=[TraceEvent.after_literal.to_ast(), self._get_copy_id_ast(orig_node_value)],
                keywords=fast.kwargs(obj=node.value),
            )
        return node



# -*- coding: utf-8 -*-
import ast
from collections import defaultdict
from enum import Enum
import logging
from typing import cast, TYPE_CHECKING
import weakref

from nbsafety.data_model.update_protocol import UpdateProtocol

if TYPE_CHECKING:
    from typing import Any, Dict, Optional, Set, Union
    import ast
    from nbsafety.safety import NotebookSafety
    from nbsafety.data_model.scope import Scope, NamespaceScope

logger = logging.getLogger(__name__)


class DataSymbolType(Enum):
    DEFAULT = 'default'
    SUBSCRIPT = 'subscript'
    FUNCTION = 'function'
    CLASS = 'class'
    IMPORT = 'import'


class DataSymbol(object):
    def __init__(
            self,
            name: 'Union[str, int]',
            symbol_type: 'DataSymbolType',
            obj: 'Any',
            containing_scope: 'Scope',
            safety: 'NotebookSafety',
            stmt_node: 'Optional[ast.AST]' = None,
            parents: 'Optional[Set[DataSymbol]]' = None,
            refresh_cached_obj=False,
            implicit=False,
    ):
        # print(containing_scope, name, obj, is_subscript)
        self.name = name
        self.symbol_type = symbol_type
        tombstone, obj_ref, has_weakref = self._update_obj_ref_inner(obj)
        self._tombstone = tombstone
        self._obj_ref = obj_ref
        self._has_weakref = has_weakref
        self.cached_obj_ref = None
        self._cached_has_weakref = None
        self.cached_obj_id = None
        self.cached_obj_type = None
        if refresh_cached_obj:
            self._refresh_cached_obj()
        self.containing_scope = containing_scope
        self.safety = safety
        self.stmt_node = self.update_stmt_node(stmt_node)
        self._funcall_live_symbols = None
        if parents is None:
            parents = set()
        self.parents: Set[DataSymbol] = parents
        self.children_by_cell_position: Dict[int, Set[DataSymbol]] = defaultdict(set)

        self.call_scope: Optional[Scope] = None
        if self.is_function:
            self.call_scope = self.containing_scope.make_child_scope(self.name)

        self.defined_cell_num = self.safety.cell_counter()

        # The notebook cell number this is required to have to not be considered stale
        self.required_cell_num = self.defined_cell_num

        self.fresher_ancestors: Set[DataSymbol] = set()
        self.namespace_stale_symbols: Set[DataSymbol] = set()

        # if implicitly created by attrsub access
        self._implicit = implicit

        # Will never be stale if no_warning is True
        self.disable_warnings = False

        self.safety.aliases[id(obj)].add(self)

    def __repr__(self) -> str:
        return f'<{self.readable_name}>'

    def __str__(self) -> str:
        return self.readable_name

    def __hash__(self):
        return hash(self.full_path)

    @property
    def readable_name(self) -> str:
        return self.containing_scope.make_namespace_qualified_name(self)

    @property
    def is_subscript(self):
        return self.symbol_type == DataSymbolType.SUBSCRIPT

    @property
    def is_class(self):
        return self.symbol_type == DataSymbolType.CLASS

    @property
    def is_function(self):
        return self.symbol_type == DataSymbolType.FUNCTION

    @property
    def is_import(self):
        return self.symbol_type == DataSymbolType.IMPORT

    @property
    def is_implicit(self):
        return self._implicit

    def _get_obj(self) -> 'Any':
        if self._has_weakref:
            return self._obj_ref()
        else:
            return self._obj_ref

    def _get_cached_obj(self) -> 'Any':
        if self._cached_has_weakref:
            return self.cached_obj_ref()
        else:
            return self.cached_obj_ref

    def shallow_clone(self, new_obj, new_containing_scope, symbol_type):
        return self.__class__(self.name, symbol_type, new_obj, new_containing_scope, self.safety)

    @property
    def obj_id(self):
        return id(self._get_obj())

    @property
    def obj_type(self):
        return type(self._get_obj())

    @property
    def namespace(self):
        return self.safety.namespaces.get(self.obj_id, None)

    @property
    def full_path(self):
        return self.containing_scope.full_path + (self.name,)

    @property
    def full_namespace_path(self):
        return self.containing_scope.make_namespace_qualified_name(self)

    @property
    def is_garbage(self):
        return (
            self._tombstone
            or self.containing_scope.is_garbage
            or not self.containing_scope.is_globally_accessible
            or (self._has_weakref and self._get_obj() is None)
        )

    @property
    def is_globally_accessible(self):
        return self.containing_scope.is_globally_accessible

    def _obj_reference_expired_callback(self, *_):
        # just write a tombstone here; we'll do a batch collect after the main part of the cell is done running
        # can potentially support GC in the background further down the line
        self._tombstone = True

    def collect_self_garbage(self):
        for parent in self.parents:
            for parent_children in parent.children_by_cell_position.values():
                parent_children.discard(self)
        for self_children in self.children_by_cell_position.values():
            for child in self_children:
                child.parents.discard(self)
        # kill the alias but leave the namespace
        # namespace needs to stick around to properly handle the staleness propagation protocol
        self._handle_aliases(readd=False)

    # def update_type(self, new_type):
    #     self.symbol_type = new_type
    #     if self.is_function:
    #         self.call_scope = self.containing_scope.make_child_scope(self.name)
    #     else:
    #         self.call_scope = None

    def update_obj_ref(self, obj, refresh_cached=True):
        tombstone, obj_ref, has_weakref = self._update_obj_ref_inner(obj)
        self._tombstone = tombstone
        self._obj_ref = obj_ref
        self._has_weakref = has_weakref
        if self.cached_obj_id is not None and self.cached_obj_id != self.obj_id:
            old_ns = self.safety.namespaces.get(self.cached_obj_id, None)
            if old_ns is not None:
                old_ns.update_obj_ref(obj)
            self._handle_aliases()
        if refresh_cached:
            self._refresh_cached_obj()

    def _handle_aliases(self, readd=True):
        old_aliases = self.safety.aliases.get(self.cached_obj_id, None)
        if old_aliases is not None:
            old_aliases.discard(self)
            if len(old_aliases) == 0:
                del self.safety.aliases[self.cached_obj_id]
        if readd:
            self.safety.aliases[self.obj_id].add(self)

    def _update_obj_ref_inner(self, obj):
        tombstone = False
        try:
            obj_ref = weakref.ref(obj, self._obj_reference_expired_callback)
            has_weakref = True
        except TypeError:
            obj_ref = obj
            has_weakref = False
        return tombstone, obj_ref, has_weakref

    def update_stmt_node(self, stmt_node):
        self.stmt_node = stmt_node
        self._funcall_live_symbols = None
        if self.is_function:
            self.safety.statement_to_func_cell[id(stmt_node)] = self
        return stmt_node

    def _refresh_cached_obj(self):
        self.cached_obj_ref = self._obj_ref
        self.cached_obj_id = self.obj_id
        self.cached_obj_type = self.obj_type
        self._cached_has_weakref = self._has_weakref

    def get_call_args(self):
        # TODO: handle lambda, objects w/ __call__, etc
        args = set()
        if self.is_function:
            assert isinstance(self.stmt_node, ast.FunctionDef)
            for arg in self.stmt_node.args.args + self.stmt_node.args.kwonlyargs:
                args.add(arg.arg)
            if self.stmt_node.args.vararg is not None:
                args.add(self.stmt_node.args.vararg.arg)
            if self.stmt_node.args.kwarg is not None:
                args.add(self.stmt_node.args.kwarg.arg)
        return args

    def create_symbols_for_call_args(self):
        for arg in self.get_call_args():
            # TODO: ideally we should try to pass the object here
            self.call_scope.upsert_data_symbol_for_name(arg, None, set(), self.stmt_node, False, propagate=False)

    @property
    def is_stale(self):
        if self.disable_warnings:
            return False
        return self.defined_cell_num < self.required_cell_num or len(self.namespace_stale_symbols) > 0

    def should_mark_stale(self, updated_dep):
        if self.disable_warnings:
            return False
        if updated_dep is self:
            return False
        return True

    def update_deps(
            self, new_deps: 'Set[DataSymbol]', overwrite=True, mutated=False, propagate=True
    ):
        # skip updates for imported symbols
        if self.is_import:
            return
        # if we get here, no longer implicit
        self._implicit = False
        # quick last fix to avoid overwriting if we appear inside the set of deps to add
        overwrite = overwrite and self not in new_deps
        new_deps.discard(self)
        if overwrite:
            for parent in self.parents - new_deps:
                for parent_children in parent.children_by_cell_position.values():
                    parent_children.discard(self)
            self.parents = set()

        for new_parent in new_deps - self.parents:
            if new_parent is None:
                continue
            new_parent.children_by_cell_position[self.safety.active_cell_position_idx].add(self)
            self.parents.add(new_parent)
        self.required_cell_num = -1
        UpdateProtocol(self.safety, self, new_deps, mutated)(propagate=propagate)
        self._refresh_cached_obj()
        self.safety.updated_symbols.add(self)

    def refresh(self: 'DataSymbol'):
        self.fresher_ancestors = set()
        self.defined_cell_num = self.safety.cell_counter()
        self.namespace_stale_symbols = set()

    def _propagate_refresh_to_namespace_parents(self, seen: 'Set[DataSymbol]'):
        if self in seen:
            return
        # print('refresh propagate', self)
        seen.add(self)
        for self_alias in self.safety.aliases[self.obj_id]:
            containing_scope: 'NamespaceScope' = cast('NamespaceScope', self_alias.containing_scope)
            if not containing_scope.is_namespace_scope:
                continue
            # if containing_scope.max_defined_timestamp == self.safety.cell_counter():
            #     return
            containing_scope.max_defined_timestamp = self.safety.cell_counter()
            containing_namespace_obj_id = containing_scope.obj_id
            # print('containing namespaces:', self.safety.aliases[containing_namespace_obj_id])
            for alias in self.safety.aliases[containing_namespace_obj_id]:
                alias.namespace_stale_symbols.discard(self)
                if not alias.is_stale:
                    alias.defined_cell_num = self.safety.cell_counter()
                    alias.fresher_ancestors = set()
                # print('working on', alias, '; stale?', alias.is_stale, alias.namespace_stale_symbols)
                alias._propagate_refresh_to_namespace_parents(seen)



# -*- coding: utf-8 -*-

# ref: https://gist.github.com/golobor/397b5099d42da476a4e6
class DotDict(dict):
    """A dict with dot access and autocompletion.

    The idea and most of the code was taken from
    http://stackoverflow.com/a/23689767,
    http://code.activestate.com/recipes/52308-the-simple-but-handy-collector-of-a-bunch-of-named/
    http://stackoverflow.com/questions/2390827/how-to-properly-subclass-dict-and-override-get-set
    """
    def __init__(self, *a, **kw):
        dict.__init__(self)
        self.update(*a, **kw)
        self.__dict__ = self

    def __setattr__(self, key, value):
        if key in dict.__dict__:
            raise AttributeError('This key is reserved for the dict methods.')
        dict.__setattr__(self, key, value)

    def __setitem__(self, key, value):
        if key in dict.__dict__:
            raise AttributeError('This key is reserved for the dict methods.')
        dict.__setitem__(self, key, value)

    def update(self, *args, **kwargs):
        for k, v in dict(*args, **kwargs).items():
            self[k] = v

    def __getstate__(self):
        return self

    def __setstate__(self, state):
        self.update(state)
        self.__dict__ = self



# -*- coding: utf-8 -*-
import ast
import builtins
from contextlib import contextmanager
import itertools
import logging
import sys
from typing import cast, TYPE_CHECKING

import astunparse

from nbsafety.analysis.attr_symbols import AttrSubSymbolChain, GetAttrSubSymbols
from nbsafety.data_model.data_symbol import DataSymbol, DataSymbolType
from nbsafety.data_model.scope import NamespaceScope
from nbsafety.tracing.mutation_event import MutationEvent
from nbsafety.tracing.recovery import on_exception_default_to, return_arg_at_index, return_val
from nbsafety.tracing.trace_events import TraceEvent, EMIT_EVENT
from nbsafety.tracing.trace_stmt import TraceStatement

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
    from types import FrameType
    from nbsafety.data_model.scope import Scope
    from nbsafety.safety import NotebookSafety
    SymbolRef = Union[str, AttrSubSymbolChain]
    AttrSubVal = Union[str, int]
    RecordedArg = Tuple[AttrSubSymbolChain, int]
    RecordedArgs = Set[RecordedArg]
    DeepRef = Tuple[int, Optional[str], Tuple[RecordedArg, ...]]
    Mutation = Tuple[int, Tuple[RecordedArg, ...], MutationEvent]
    RefCandidate = Optional[Tuple[int, int, Optional[str]]]
    DeepRefCandidate = Tuple[RefCandidate, MutationEvent, RecordedArgs]
    SavedStoreData = Tuple[NamespaceScope, Any, AttrSubVal, bool]
    LexicalCallNestingStack = List[Scope]


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


ARG_MUTATION_EXCEPTED_MODULES = {
    'alt',
    'altair',
    'display',
    'logging',
    'matplotlib',
    'pyplot',
    'plot',
    'plt',
    'seaborn',
    'sns',
    'widget',
}


class ListLiteral(list):
    pass


class DictLiteral(dict):
    pass


def _make_weakrefable_literal(literal):
    if type(literal) == list:
        return ListLiteral(literal)
    elif type(literal) == dict:
        return DictLiteral(literal)
    else:
        return literal


def _finish_tracing_reset():
    # do nothing; we just want to trigger the newly reenabled tracer with a 'call' event
    pass


class TracingManager(object):
    def __init__(self, safety: 'NotebookSafety'):
        self.safety = safety
        self.trace_event_counter = 0
        self.prev_event: Optional[TraceEvent] = None
        self.prev_trace_stmt: Optional[TraceStatement] = None
        self.seen_stmts: 'Set[int]' = set()
        self.call_depth = 0
        self.traced_statements: Dict[int, TraceStatement] = {}
        self.tracing_enabled = False
        self.tracing_reset_pending = False

        setattr(builtins, EMIT_EVENT, self._emit_event)

        self._stack: 'List[Tuple[Any, ...]]' = []
        self._stack_item_initializers: 'Dict[str, Callable[[], Any]]' = {}
        self._stack_items_with_manual_initialization: 'Set[str]' = set()
        self._registering_stack_state_context = False
        with self._register_stack_state():
            # everything here should be copyable
            self.prev_trace_stmt_in_cur_frame: Optional[TraceStatement] = None
            self.loaded_data_symbols: Set[DataSymbol] = set()
            self.saved_store_data: List[SavedStoreData] = []
            self.deep_refs: Set[DeepRef] = set()
            self.mutations: Set[Mutation] = set()
            self.deep_ref_candidates: List[DeepRefCandidate] = []
            self.nested_call_stack: LexicalCallNestingStack = []
            self.should_record_args = False
            self.should_record_args_stack: List[bool] = []
            self.literal_namespace: Optional[NamespaceScope] = None
            self.first_obj_id_in_chain: Optional[int] = None
            with self._needing_manual_initialization():
                self.cur_frame_original_scope = safety.global_scope
                self.active_scope = safety.global_scope
                self.inside_lambda = False

    @property
    def _stack_item_names(self):
        return itertools.chain(self._stack_item_initializers.keys(), self._stack_items_with_manual_initialization)

    @contextmanager
    def _register_stack_state(self):
        self._registering_stack_state_context = True
        original_state = set(self.__dict__.keys())
        yield
        self._registering_stack_state_context = False
        stack_item_names = set(self.__dict__.keys() - original_state)
        for stack_item_name in stack_item_names - self._stack_items_with_manual_initialization:
            stack_item = self.__dict__[stack_item_name]
            if stack_item is None:
                self._stack_item_initializers[stack_item_name] = lambda: None
            elif isinstance(stack_item, bool):
                init_val = bool(stack_item)
                self._stack_item_initializers[stack_item_name] = lambda: init_val
            else:
                self._stack_item_initializers[stack_item_name] = type(stack_item)

    @contextmanager
    def _needing_manual_initialization(self):
        assert self._registering_stack_state_context
        original_state = set(self.__dict__.keys())
        yield
        self._stack_items_with_manual_initialization = set(self.__dict__.keys() - original_state)

    @contextmanager
    def _push_stack(self):
        self._stack.append(tuple(self.__dict__[stack_item] for stack_item in self._stack_item_names))
        for stack_item, initializer in self._stack_item_initializers.items():
            self.__dict__[stack_item] = initializer()
        for stack_item in self._stack_items_with_manual_initialization:
            del self.__dict__[stack_item]
        yield
        uninitialized_items = []
        for stack_item in self._stack_items_with_manual_initialization:
            if stack_item not in self.__dict__:
                uninitialized_items.append(stack_item)
        if len(uninitialized_items) > 0:
            raise ValueError(
                "Stack item(s) %s requiring manual initialization were not initialized" % uninitialized_items
            )

    def _handle_call_transition(self, trace_stmt: 'TraceStatement'):
        new_scope = trace_stmt.get_post_call_scope()
        with self._push_stack():
            # TODO: figure out a better way to determine if we're inside a lambda
            #  could this one lead to a false negative if a lambda is in the default of a function def kwarg?
            self.inside_lambda = not isinstance(
                trace_stmt.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)
            )
            self.cur_frame_original_scope = new_scope
            self.active_scope = new_scope
        self.prev_trace_stmt_in_cur_frame = self.prev_trace_stmt = trace_stmt

    def _pop_stack(self):
        for stack_item_name, stack_item in zip(self._stack_item_names, self._stack.pop()):
            self.__dict__[stack_item_name] = stack_item

    def _check_prev_stmt_done_executing_hook(self, event: 'TraceEvent', trace_stmt: 'TraceStatement'):
        if event == TraceEvent.after_stmt:
            trace_stmt.finished_execution_hook()
        elif event == TraceEvent.return_ and self.prev_event not in (TraceEvent.call, TraceEvent.exception):
            # ensuring prev != call ensures we're not inside of a stmt with multiple calls (such as map w/ lambda)
            if self.prev_trace_stmt is not None:
                self.prev_trace_stmt.finished_execution_hook()
            # prev_overall = self.prev_trace_stmt
            # if prev_overall is not None and prev_overall is not self._stack[-1][0]:
            #     # this condition ensures we're not inside of a stmt with multiple calls (such as map w/ lambda)
            #     prev_overall.finished_execution_hook()

    def _handle_return_transition(self, trace_stmt: 'TraceStatement'):
        inside_lambda = self.inside_lambda
        cur_frame_scope = self.cur_frame_original_scope
        self._pop_stack()
        return_to_stmt = self.prev_trace_stmt_in_cur_frame
        assert return_to_stmt is not None
        if self.prev_event != TraceEvent.exception:
            # exception events are followed by return events until we hit an except clause
            # no need to track dependencies in this case
            if isinstance(return_to_stmt.stmt_node, ast.ClassDef):
                return_to_stmt.class_scope = cast(NamespaceScope, cur_frame_scope)
            elif isinstance(trace_stmt.stmt_node, ast.Return) or inside_lambda:
                if not trace_stmt.lambda_call_point_deps_done_once:
                    trace_stmt.lambda_call_point_deps_done_once = True
                    return_to_stmt.call_point_deps.append(trace_stmt.compute_rval_dependencies())

    def state_transition_hook(
            self,
            event: 'TraceEvent',
            trace_stmt: 'TraceStatement'
    ):
        self.trace_event_counter += 1

        self._check_prev_stmt_done_executing_hook(event, trace_stmt)

        if event == TraceEvent.call:
            self._handle_call_transition(trace_stmt)
        if event == TraceEvent.return_:
            self._handle_return_transition(trace_stmt)
        self.prev_event = event

    def _emit_event(self, evt: str, orig_node_id: int, **kwargs: 'Any'):
        event = TraceEvent(evt)
        if event == TraceEvent.before_stmt:
            return self.before_stmt_tracer(orig_node_id, sys._getframe().f_back)
        elif event == TraceEvent.after_stmt:
            return self.after_stmt_tracer(orig_node_id, sys._getframe().f_back, ret_expr=kwargs.get('ret_expr', None))
        elif event in (TraceEvent.attribute, TraceEvent.subscript):
            return self.attrsub_tracer(
                kwargs['obj'],
                kwargs['attr_or_sub'],
                kwargs['ctx'],
                kwargs['call_context'],
                is_subscript=event == TraceEvent.subscript,
                obj_name=kwargs.get('name', None),
            )
        elif event == TraceEvent.after_attrsub_chain:
            return self.end_tracer(kwargs['obj'], kwargs['call_context'])
        elif event == TraceEvent.argument:
            return self.arg_recorder(kwargs['obj'], self.safety.ast_node_by_id[orig_node_id])
        elif event == TraceEvent.before_arg_list:
            return self.before_argument_list(kwargs['obj'])
        elif event == TraceEvent.after_arg_list:
            return self.after_argument_list(kwargs['obj'], kwargs['is_attrsub'], kwargs['inside_chain'])
        elif event == TraceEvent.before_literal:
            pass
        elif event == TraceEvent.after_literal:
            return self.literal_tracer(kwargs['obj'])
        else:
            raise ValueError('Unsupported event: %s' % event)

    def _get_namespace_for_obj(self, obj: 'Any', obj_name: 'Optional[str]' = None) -> 'NamespaceScope':
        obj_id = id(obj)
        ns = self.safety.namespaces.get(obj_id, None)
        # print('%s attrsub %s of obj %s' % (ctx, attr_or_subscript, obj))
        if ns is not None:
            return ns
        class_scope = self.safety.namespaces.get(id(obj.__class__), None)
        if class_scope is not None:
            # print('found class scope %s containing %s' % (class_scope, list(class_scope.all_data_symbols_this_indentation())))
            ns = class_scope.clone(obj)
            if obj_name is not None:
                ns.scope_name = obj_name
        else:
            # print('no scope for class', obj.__class__)
            # if self.prev_trace_stmt.finished:
            #     # avoid creating new scopes if we already did this computation
            #     self.active_scope = None
            #     return obj
            try:
                scope_name = next(iter(self.safety.aliases.get(obj_id, None))).name if obj_name is None else obj_name
            except (TypeError, StopIteration):
                scope_name = '<unknown namespace>'
            ns = NamespaceScope(obj, self.safety, scope_name, parent_scope=None)
        # FIXME: brittle strategy for determining parent scope of obj
        if ns.parent_scope is None:
            if (
                    obj_name is not None and
                    obj_name not in self.prev_trace_stmt_in_cur_frame.frame.f_locals
            ):
                parent_scope = self.safety.global_scope
            else:
                parent_scope = self.active_scope
            ns.parent_scope = parent_scope
        return ns

    @on_exception_default_to(return_arg_at_index(1, logger))
    def attrsub_tracer(
            self, obj, attr_or_subscript, ctx: str, call_context: bool, is_subscript: bool, obj_name: 'Optional[str]'
    ):
        if not self.tracing_enabled:
            return obj
        should_record_args = False
        try:
            if obj is None:
                return None
            obj_id = id(obj)
            if self.first_obj_id_in_chain is None:
                self.first_obj_id_in_chain = obj_id
            if isinstance(attr_or_subscript, tuple):
                if not all(isinstance(v, (str, int)) for v in attr_or_subscript):
                    return obj
            elif not isinstance(attr_or_subscript, (str, int)):
                return obj

            scope = self._get_namespace_for_obj(obj, obj_name=obj_name)
            self.active_scope = scope
            # if scope is None:  # or self.prev_trace_stmt.finished:
            #     if ctx in ('Store', 'AugStore'):
            #         self.active_scope = self.original_active_scope
            #     return obj
            if scope is None or self.prev_trace_stmt_in_cur_frame.finished:
                return obj
            elif ctx in ('Store', 'AugStore') and scope is not None:
                self.saved_store_data.append((scope, obj, attr_or_subscript, is_subscript))
                # reset active scope here
                self.active_scope = self.cur_frame_original_scope
            if ctx == 'Load':
                # save off event counter and obj_id
                # if event counter didn't change when we process the Call retval, and if the
                # retval is None, this is a likely signal that we have a mutation
                # TODO: this strategy won't work if the arguments themselves lead to traced function calls
                # print('looking for', attr_or_subscript)
                data_sym = scope.lookup_data_symbol_by_name_this_indentation(
                    attr_or_subscript, is_subscript=is_subscript
                )
                try:
                    obj_attr_or_sub = self.safety.retrieve_namespace_attr_or_sub(
                        obj, attr_or_subscript, is_subscript
                    )
                    if data_sym is None:
                        symbol_type = DataSymbolType.SUBSCRIPT if is_subscript else DataSymbolType.DEFAULT
                        data_sym = DataSymbol(
                            attr_or_subscript,
                            symbol_type,
                            obj_attr_or_sub,
                            scope,
                            self.safety,
                            stmt_node=None,
                            parents=None,
                            refresh_cached_obj=True,
                            implicit=True,
                        )
                        # this is to prevent refs to the scope object from being considered as stale if we just load it
                        data_sym.defined_cell_num = data_sym.required_cell_num = scope.max_defined_timestamp
                        scope.put(attr_or_subscript, data_sym)
                        # print('put', data_sym, 'in', scope.full_namespace_path)
                    elif data_sym.obj_id != id(obj_attr_or_sub):
                        data_sym.update_obj_ref(obj_attr_or_sub)
                except:
                    pass
                if call_context:
                    should_record_args = True
                    mutation_event = MutationEvent.normal
                    if isinstance(obj, list) and attr_or_subscript == 'append':
                        mutation_event = MutationEvent.list_append
                    self.deep_ref_candidates.append(
                        ((self.trace_event_counter, obj_id, obj_name), mutation_event, set())
                    )
                elif data_sym is not None:
                    # TODO: if we have a.b.c, will this consider a.b loaded as well as a.b.c? This is bad if so.
                    self.loaded_data_symbols.add(data_sym)
            return obj
        finally:
            if call_context:
                self.should_record_args_stack.append(self.should_record_args)
                self.should_record_args = should_record_args

    @on_exception_default_to(return_arg_at_index(1, logger))
    def end_tracer(self, obj: 'Any', call_context: bool):
        first_obj_id_in_chain = self.first_obj_id_in_chain
        self.first_obj_id_in_chain = None
        if not self.tracing_enabled:
            return obj
        if self.prev_trace_stmt_in_cur_frame.finished:
            self.active_scope = self.cur_frame_original_scope
            return obj
        if call_context and len(self.deep_ref_candidates) > 0:
            (evt_counter, obj_id, obj_name), mutation_event, recorded_args = self.deep_ref_candidates.pop()
            if evt_counter == self.trace_event_counter:
                if obj is None:
                    if mutation_event == MutationEvent.normal:
                        try:
                            top_level_sym = next(iter(self.safety.aliases[first_obj_id_in_chain]))
                            if top_level_sym.is_import and top_level_sym.name not in ARG_MUTATION_EXCEPTED_MODULES:
                                # TODO: should it be the other way around? i.e. allow-list for arg mutations, starting
                                #  with np.random.seed?
                                for recorded_arg, _ in recorded_args:
                                    if len(recorded_arg.symbols) > 0:
                                        # only make this an arg mutation event if it looks like there's an arg to mutate
                                        mutation_event = MutationEvent.arg_mutate
                                        break
                        except:
                            pass
                    self.mutations.add((obj_id, tuple(recorded_args), mutation_event))
                else:
                    self.deep_refs.add((obj_id, obj_name, tuple(recorded_args)))
        # print('reset active scope from', self.active_scope, 'to', self.original_active_scope)
        self.active_scope = self.cur_frame_original_scope
        return obj

    @on_exception_default_to(return_arg_at_index(1, logger))
    def arg_recorder(self, arg_obj: 'Any', arg_node: 'ast.AST'):
        if not self.tracing_enabled:
            return arg_obj
        if self.prev_trace_stmt_in_cur_frame.finished or not self.should_record_args:
            return arg_obj
        if not isinstance(arg_node, (ast.Attribute, ast.Subscript, ast.Call, ast.Name)):
            return arg_obj
        if len(self.deep_ref_candidates) == 0:
            logger.error('Error: no associated symbol for recorded args; skipping recording')
            return arg_obj

        arg_obj_id = id(arg_obj)
        # TODO: we should be able to get the actual data symbol during live tracing,
        #  instead of trying to resolve from an attrsub chain determined via analysis
        recorded_arg = GetAttrSubSymbols()(arg_node)
        self.deep_ref_candidates[-1][-1].add((recorded_arg, arg_obj_id))

        return arg_obj

    @on_exception_default_to(return_arg_at_index(1, logger))
    def before_argument_list(self, obj):
        if not self.tracing_enabled:
            return obj
        # if self.prev_trace_stmt.finished:
        #     return obj
        self.nested_call_stack.append(self.active_scope)
        self.active_scope = self.cur_frame_original_scope
        return obj

    @on_exception_default_to(return_arg_at_index(1, logger))
    def after_argument_list(self, obj: 'Any', should_pop_should_record_args_stack: bool, inside_chain: bool):
        if not self.tracing_enabled:
            return obj
        # if self.prev_trace_stmt.finished:
        #     return obj
        self.active_scope = self.nested_call_stack.pop()
        if should_pop_should_record_args_stack:
            self.should_record_args = self.should_record_args_stack.pop()
        if inside_chain:
            # TODO: I don't think any test exercises this atm
            self.active_scope = self._get_namespace_for_obj(obj)
        return obj

    @on_exception_default_to(return_arg_at_index(1, logger))
    def literal_tracer(self, literal):
        literal = _make_weakrefable_literal(literal)
        if not self.tracing_enabled:
            return literal
        if self.prev_trace_stmt_in_cur_frame.finished:
            return literal
        if isinstance(literal, (dict, list, tuple)):
            scope = NamespaceScope(
                literal, self.safety, None, self.prev_trace_stmt_in_cur_frame.scope
            )
            gen = literal.items() if isinstance(literal, dict) else enumerate(literal)
            for i, obj in gen:
                scope.upsert_data_symbol_for_name(
                    i, obj, set(), self.prev_trace_stmt_in_cur_frame.stmt_node, True
                )
            self.literal_namespace = scope
        return literal

    def after_stmt_tracer(self, stmt_id: int, frame: 'FrameType', ret_expr: 'Optional[Any]' = None):
        if stmt_id in self.seen_stmts:
            return ret_expr
        stmt = self.safety.ast_node_by_id.get(stmt_id, None)
        if stmt is not None:
            self._sys_tracer(frame, TraceEvent.after_stmt, stmt)
        return ret_expr

    def before_stmt_tracer(self, stmt_id: int, frame: 'FrameType'):
        if stmt_id in self.seen_stmts:
            return
        # logger.warning('reenable tracing: %s', site_id)
        if self.prev_trace_stmt_in_cur_frame is not None:
            prev_trace_stmt_in_cur_frame = self.prev_trace_stmt_in_cur_frame
            # both of the following stmts should be processed when body is entered
            if isinstance(prev_trace_stmt_in_cur_frame.stmt_node, (ast.For, ast.If, ast.With)):
                self.after_stmt_tracer(prev_trace_stmt_in_cur_frame.stmt_id, frame)
        trace_stmt = self.traced_statements.get(stmt_id, None)
        if trace_stmt is None:
            trace_stmt = TraceStatement(
                self.safety,
                frame,
                cast(ast.stmt, self.safety.ast_node_by_id[stmt_id]),
                self.cur_frame_original_scope
            )
            self.traced_statements[stmt_id] = trace_stmt
        self.prev_trace_stmt_in_cur_frame = trace_stmt
        self.prev_trace_stmt = trace_stmt
        if not self.tracing_enabled:
            assert not self.tracing_reset_pending
            self._enable_tracing()
            self.tracing_reset_pending = True
            _finish_tracing_reset()  # trigger the tracer with a frame

    def after_stmt_reset_hook(self):
        self.loaded_data_symbols.clear()
        self.saved_store_data.clear()
        self.deep_refs.clear()
        self.mutations.clear()
        self.deep_ref_candidates.clear()
        self.nested_call_stack.clear()
        self.should_record_args_stack.clear()
        self.active_scope = self.cur_frame_original_scope
        self.should_record_args = False
        self.literal_namespace = None
        self.first_obj_id_in_chain = None

    def _enable_tracing(self):
        assert not self.tracing_enabled
        self.tracing_enabled = True
        sys.settrace(self._sys_tracer)

    def _disable_tracing(self, check_enabled=True):
        if check_enabled:
            assert self.tracing_enabled
        self.tracing_enabled = False
        sys.settrace(None)

    @contextmanager
    def tracing_context(self):
        try:
            self._enable_tracing()
            yield
        finally:
            self._disable_tracing(check_enabled=False)

    def _attempt_to_reenable_tracing(self, frame: 'FrameType') -> None:
        if self.safety.is_develop:
            assert self.tracing_reset_pending, 'expected tracing reset to be pending!'
            assert self.call_depth > 0, 'expected managed call depth > 0, got %d' % self.call_depth
        self.tracing_reset_pending = False
        call_depth = 0
        while frame is not None:
            if frame.f_code.co_filename.startswith('<ipython-input'):
                call_depth += 1
            frame = frame.f_back
        if self.safety.is_develop:
            assert call_depth >= 1, 'expected call depth >= 1, got %d' % call_depth
        # TODO: allow reenabling tracing beyond just at the top level
        if call_depth != 1:
            self._disable_tracing()
            return
        # at this point, we can be sure we're at the top level
        # because tracing was enabled in a handler and not in the
        # top level, we need to clear the stack, since we won't
        # catch the return event
        self.call_depth = 0
        self._stack = []
        self.nested_call_stack = []
        if self.safety.config.trace_messages_enabled:
            logger.warning('reenable tracing >>>')

    @on_exception_default_to(return_val(None, logger))
    def _sys_tracer(self, frame: 'FrameType', evt: 'Union[str, TraceEvent]', extra):
        if isinstance(evt, str):
            event = TraceEvent(evt)
        else:
            event = evt

        if self.tracing_reset_pending:
            assert event == TraceEvent.call
            self._attempt_to_reenable_tracing(frame)
            return None

        # notebook cells have filenames that appear as '<ipython-input...>'
        if frame.f_code.co_filename.startswith('<ipython-input'):
            self.safety.maybe_set_name_to_cell_num_mapping(frame)
        else:
            return None

        if event == TraceEvent.line:
            return self._sys_tracer

        if event not in (TraceEvent.return_, TraceEvent.after_stmt) and not self.tracing_enabled:
            logger.warning('skip %s', event)
            return None

        # IPython quirk -- every line in outer scope apparently wrapped in lambda
        # We want to skip the outer 'call' and 'return' for these
        if event == TraceEvent.call:
            self.call_depth += 1
            if self.call_depth == 1:
                return self._sys_tracer

        if event == TraceEvent.return_:
            self.call_depth -= 1
            if self.call_depth == 0:
                return self._sys_tracer

        cell_num, lineno = self.safety.get_position(frame)

        if event == TraceEvent.after_stmt:
            stmt_node = extra
        else:
            try:
                stmt_node = self.safety.statement_cache[cell_num][lineno]
            except KeyError:
                if self.safety.is_develop:
                    logger.warning("got key error for stmt node in cell %d, line %d", cell_num, lineno)
                return self._sys_tracer

        trace_stmt = self.traced_statements.get(id(stmt_node), None)
        if trace_stmt is None:
            trace_stmt = TraceStatement(self.safety, frame, stmt_node, self.cur_frame_original_scope)
            self.traced_statements[id(stmt_node)] = trace_stmt

        if self.safety.config.trace_messages_enabled:
            codeline = astunparse.unparse(stmt_node).strip('\n').split('\n')[0]
            codeline = ' ' * getattr(stmt_node, 'col_offset', 0) + codeline
            logger.warning(' %3d: %10s >>> %s', lineno, event, codeline)
        if event == TraceEvent.call:
            if trace_stmt.call_seen:
                if self.safety.config.trace_messages_enabled:
                    logger.warning(' disable tracing >>>')
                self._disable_tracing()
                return None
            trace_stmt.call_seen = True
        self.state_transition_hook(event, trace_stmt)
        return self._sys_tracer



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

import versioneer

pkg_name = 'nbsafety'


def read_file(fname):
    with open(fname, 'r', encoding='utf8') as f:
        return f.read()


history = read_file('HISTORY.rst')
requirements = read_file('requirements.txt').strip().split()

setup(
    name=pkg_name,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author='Stephen Macke',
    author_email='stephen.macke@gmail.com',
    description='Fearless interactivity for Jupyter notebooks.',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/nbsafety-project/nbsafety',
    packages=find_packages(exclude=[
        'binder',
        'docs',
        'scratchspace',
        'notebooks',
        'img',
        'test',
        'scripts',
        'markdown',
        'versioneer.py',
        'frontend',
        'blueprint.json',
    ]),
    include_package_data=True,
    data_files=[
        # like `jupyter nbextension install --sys-prefix`
        ("share/jupyter/nbextensions/nbsafety", [
            "nbsafety/resources/nbextension/index.js",
        ]),
        ("share/jupyter/nbextensions/nbsafety", [
            "nbsafety/resources/nbextension/index.js.map",
        ]),
        # like `jupyter nbextension enable --sys-prefix`
        ("etc/jupyter/nbconfig/notebook.d", [
            "nbsafety/resources/nbextension/nbsafety.json",
        ]),
        # like `python -m nbsafety.install --sys-prefix`
        ("share/jupyter/kernels/nbsafety", [
            "nbsafety/resources/kernel/kernel.json",
        ]),
        ("share/jupyter/kernels/nbsafety", [
            "nbsafety/resources/kernel/logo-32x32.png",
        ]),
        ("share/jupyter/kernels/nbsafety", [
            "nbsafety/resources/kernel/logo-64x64.png",
        ]),
    ],
    install_requires=requirements,
    license='BSD-3-Clause',
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)

# python setup.py sdist bdist_wheel --universal
# twine upload dist/*



#!/usr/bin/env python
#
# Copyright 2017 TWO SIGMA OPEN SOURCE, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# smacke: this was originally copied. license + copyright preserved as originally stated.
# ref: https://github.com/twosigma/beakerx/

import os
import sys
import subprocess
import signal

PORT = 9999


def kill_processes(name, port=None):
    try:
        pidlist = map(int, subprocess.check_output(['pgrep', '-f', name]).split())
    except subprocess.CalledProcessError:
        pidlist = []
    if port is not None:
        # filter on port to avoid killing other instances of jupyter server
        listening_on_port = map(int, subprocess.check_output(['lsof', '-ti', f'tcp:{PORT}']).split())
        pidlist = set(pidlist) & set(listening_on_port)
    for pid in pidlist:
        os.kill(pid, signal.SIGKILL)


# create handler for Ctrl+C
def make_signal_handler(nbsafety):
    def _signal_handler(sgnl, frame):
        os.killpg(os.getpgid(nbsafety.pid), signal.SIGKILL)
        kill_processes('webdriver')
        kill_processes('jupyter', port=PORT)
        sys.exit(20)
    return _signal_handler


def main():
    # update environment
    subprocess.call('yarn install', shell=True)

    # start jupyter notebook
    nb_command = (
        f'jupyter lab --no-browser --notebook-dir="{os.path.abspath("..")}" '
        f'--NotebookApp.token="" --port {PORT}'
    )
    nbsafety = subprocess.Popen(
        nb_command,
        shell=True,
        executable="/bin/bash",
        preexec_fn=os.setsid,
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE
    )
    # wait for notebook server to start up
    while 1:
        line = nbsafety.stdout.readline().decode('utf-8').strip()
        if not line:
            continue
        print(line)
        if 'Jupyter Notebook' in line and 'is running' in line:
            break

    signal.signal(signal.SIGINT, make_signal_handler(nbsafety))
    # start webdriverio
    result = subprocess.call('yarn run test', shell=True)

    # Send the signal to all the process groups
    os.killpg(os.getpgid(nbsafety.pid), signal.SIGKILL)

    if result:
        return 20


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
import ast
from contextlib import contextmanager
import logging
from typing import TYPE_CHECKING

from nbsafety.analysis import (
    AttrSubSymbolChain, get_symbol_edges, stmt_contains_lval
)
from nbsafety.data_model.scope import NamespaceScope
from nbsafety.tracing.mutation_event import MutationEvent

if TYPE_CHECKING:
    from types import FrameType
    from typing import List, Optional, Set
    from nbsafety.data_model.data_symbol import DataSymbol
    from nbsafety.data_model.scope import Scope
    from nbsafety.safety import NotebookSafety

logger = logging.getLogger(__name__)


class TraceStatement(object):
    def __init__(self, safety: 'NotebookSafety', frame: 'FrameType', stmt_node: 'ast.stmt', scope: 'Scope'):
        self.safety = safety
        self.frame = frame
        self.stmt_node = stmt_node
        self.scope = scope
        self.class_scope: Optional[NamespaceScope] = None
        self.call_point_deps: List[Set[DataSymbol]] = []
        self.lambda_call_point_deps_done_once = False
        self.call_seen = False

    @contextmanager
    def replace_active_scope(self, new_active_scope):
        old_scope = self.scope
        self.scope = new_active_scope
        yield
        self.scope = old_scope

    @property
    def finished(self):
        return self.stmt_id in self.safety.tracing_manager.seen_stmts

    @property
    def stmt_id(self):
        return id(self.stmt_node)

    def _contains_lval(self):
        return stmt_contains_lval(self.stmt_node)

    def compute_rval_dependencies(self, rval_symbol_refs=None):
        if rval_symbol_refs is None:
            symbol_edges, _ = get_symbol_edges(self.stmt_node)
            if len(symbol_edges) == 0:
                rval_symbol_refs = set()
            else:
                rval_symbol_refs = set.union(*symbol_edges.values()) - {None}
        rval_data_symbols = set()
        for name in rval_symbol_refs:
            if name is None:
                continue
            maybe_rval_dsym = self.scope.lookup_data_symbol_by_name(name)
            if maybe_rval_dsym is not None:
                rval_data_symbols.add(maybe_rval_dsym)
            # else:
            #     assert not isinstance(name, AttrSubSymbolChain)
        return rval_data_symbols.union(*self.call_point_deps) | self.safety.tracing_manager.loaded_data_symbols

    def get_post_call_scope(self):
        old_scope = self.safety.tracing_manager.cur_frame_original_scope
        if isinstance(self.stmt_node, ast.ClassDef):
            # classes need a new scope before the ClassDef has finished executing,
            # so we make it immediately
            return self.scope.make_child_scope(self.stmt_node.name, obj_id=-1)

        if not isinstance(self.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            # TODO: probably the right thing is to check is whether a lambda appears somewhere inside the ast node
            # if not isinstance(self.ast_node, ast.Lambda):
            #     raise TypeError('unexpected type for ast node %s' % self.ast_node)
            return old_scope
        func_name = self.stmt_node.name
        func_cell = self.safety.statement_to_func_cell.get(id(self.stmt_node), None)
        if func_cell is None:
            # TODO: brittle; assumes any user-defined and traceable function will always be present; is this safe?
            return old_scope
        if not func_cell.is_function:
            if self.safety.is_develop:
                raise TypeError('got non-function symbol %s for name %s' % (func_cell.full_path, func_name))
            else:
                # TODO: log an error to a file
                return old_scope
        if not self.finished:
            func_cell.create_symbols_for_call_args()
        return func_cell.call_scope

    def _handle_attrsub_stores(self, symbol_edges, deep_rval_deps):
        if len(symbol_edges) == 0:
            rval_deps = deep_rval_deps
        else:
            rval_deps = self.compute_rval_dependencies(
                rval_symbol_refs=set.union(*symbol_edges.values()) - {None}
            ) | deep_rval_deps
        for scope, obj, attr_or_sub, is_subscript in self.safety.tracing_manager.saved_store_data:
            try:
                attr_or_sub_obj = self.safety.retrieve_namespace_attr_or_sub(obj, attr_or_sub, is_subscript)
            except:
                continue
            should_overwrite = not isinstance(self.stmt_node, ast.AugAssign)
            scope_to_use = scope.get_earliest_ancestor_containing(id(attr_or_sub_obj), is_subscript)
            if scope_to_use is None:
                # Nobody before `scope` has it, so we'll insert it at this level
                scope_to_use = scope
            scope_to_use.upsert_data_symbol_for_name(
                attr_or_sub, attr_or_sub_obj, rval_deps, self.stmt_node, is_subscript,
                overwrite=should_overwrite, is_function_def=False, class_scope=None
            )
            # print(scope_to_use, 'upsert', attr_or_sub, attr_or_sub_obj, rval_deps)
            if len(self.safety.tracing_manager.saved_store_data) == 1:
                break
        else:
            return None, None

        return scope_to_use, attr_or_sub

    def _handle_literal_namespace(self, lval_name, rval_names, stored_attrsub_scope, stored_attrsub_name):
        # remaining_rval_names = set(rval_names)
        remaining_rval_names = rval_names
        if self.safety.tracing_manager.literal_namespace is None:
            return remaining_rval_names
        literal_namespace = self.safety.tracing_manager.literal_namespace
        self.safety.tracing_manager.literal_namespace = None
        if lval_name is None:
            if stored_attrsub_name is None:
                literal_namespace.scope_name = '<unknown namespace>'
            else:
                literal_namespace.scope_name = stored_attrsub_name
                if stored_attrsub_scope is not None:
                    literal_namespace.parent_scope = stored_attrsub_scope
        else:
            literal_namespace.scope_name = lval_name
        self.safety.namespaces[literal_namespace.obj_id] = literal_namespace

        # TODO: need tighter integration w/ assignment edges to allow for accurate drawing of edges to literal elements
        return remaining_rval_names
        # if len(rval_names) != literal_namespace.num_subscript_symbols:
        #     return remaining_rval_names
        #
        # # FIXME: rval_names can be traversed in the wrong order!
        # for rval_name, literal_namespace_symbol in zip(
        #         rval_names, literal_namespace.all_data_symbols_this_indentation(exclude_class=True, is_subscript=True)
        # ):
        #     if rval_name is None or rval_name == lval_name:
        #         continue
        #     literal_namespace_sym_parent = self.scope.lookup_data_symbol_by_name(rval_name)
        #     if literal_namespace_sym_parent is None:
        #         continue
        #     remaining_rval_names.discard(rval_name)
        #     literal_namespace_sym_parent.children.add(literal_namespace_symbol)
        #     literal_namespace_symbol.parents.add(literal_namespace_sym_parent)
        #
        # return remaining_rval_names

    def _make_lval_data_symbols(self):
        symbol_edges, should_overwrite = get_symbol_edges(self.stmt_node)
        deep_rval_deps = self._gather_deep_ref_rval_dsyms()
        is_function_def = isinstance(self.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef))
        is_class_def = isinstance(self.stmt_node, ast.ClassDef)
        is_import = isinstance(self.stmt_node, (ast.Import, ast.ImportFrom))
        if is_function_def or is_class_def:
            assert len(symbol_edges) == 1
            # assert not lval_symbol_refs.issubset(rval_symbol_refs)

        stored_attrsub_scope, stored_attrsub_name = self._handle_attrsub_stores(symbol_edges, deep_rval_deps)
        for lval_name, rval_names in symbol_edges.items():
            rval_names = self._handle_literal_namespace(
                lval_name, rval_names, stored_attrsub_scope, stored_attrsub_name
            )
            if lval_name is None:
                continue

            should_overwrite_for_name = should_overwrite and lval_name not in rval_names
            rval_deps = self.compute_rval_dependencies(rval_symbol_refs=rval_names - {lval_name}) | deep_rval_deps
            # print('create edges from', rval_deps, 'to', lval_name, should_overwrite_for_name)
            if is_class_def:
                assert self.class_scope is not None
                class_ref = self.frame.f_locals[self.stmt_node.name]
                class_obj_id = id(class_ref)
                self.class_scope.obj_id = class_obj_id
                self.safety.namespaces[class_obj_id] = self.class_scope
            # if is_function_def:
            #     print('create function', name, 'in scope', self.scope)
            try:
                obj = self.frame.f_locals[lval_name]
                self.scope.upsert_data_symbol_for_name(
                    lval_name, obj, rval_deps, self.stmt_node, False,
                    overwrite=should_overwrite_for_name, is_function_def=is_function_def, is_import=is_import,
                    class_scope=self.class_scope, propagate=not isinstance(self.stmt_node, ast.For)
                )
            except KeyError:
                logger.warning('keyerror for %s', lval_name)
                pass

    def _gather_deep_ref_rval_dsyms(self):
        deep_ref_rval_dsyms = set()
        for deep_ref_obj_id, deep_ref_name, deep_ref_args in self.safety.tracing_manager.deep_refs:
            deep_ref_arg_dsyms = set()
            for arg in deep_ref_args:
                if isinstance(arg, str):
                    deep_ref_arg_dsyms.add(self.scope.lookup_data_symbol_by_name(arg))
                elif isinstance(arg, AttrSubSymbolChain):
                    deep_ref_arg_dsyms.add(self.scope.get_most_specific_data_symbol_for_attrsub_chain(arg)[0])
            deep_ref_arg_dsyms.discard(None)
            deep_ref_rval_dsyms |= deep_ref_arg_dsyms
            if deep_ref_name is None:
                deep_ref_rval_dsyms |= self.safety.aliases.get(deep_ref_obj_id, set())
            else:
                deep_ref_dc = self.scope.lookup_data_symbol_by_name(deep_ref_name)
                if deep_ref_dc is not None and deep_ref_dc.obj_id == deep_ref_obj_id:
                    deep_ref_rval_dsyms.add(deep_ref_dc)
                else:
                    deep_ref_rval_dsyms |= self.safety.aliases.get(deep_ref_obj_id, set())
        return deep_ref_rval_dsyms

    def handle_dependencies(self):
        if not self.safety.dependency_tracking_enabled:
            return
        for mutated_obj_id, mutation_args, mutation_event in self.safety.tracing_manager.mutations:
            if mutation_event == MutationEvent.arg_mutate:
                for _, arg_id in mutation_args:
                    for mutated_sym in self.safety.aliases[arg_id]:
                        # TODO: happens when module mutates args
                        #  should we add module as a dep in this case?
                        mutated_sym.update_deps(set(), overwrite=False, mutated=True)
                continue

            mutation_arg_dsyms = set()
            for arg, _ in mutation_args:
                mutation_arg_dsyms.add(self.scope.get_most_specific_data_symbol_for_attrsub_chain(arg)[0])
            mutation_arg_dsyms.discard(None)

            # NOTE: this next block is necessary to ensure that we add the argument as a namespace child
            # of the mutated symbol. This helps to avoid propagating through to dependency children that are
            # themselves namespace children.
            if mutation_event == MutationEvent.list_append and len(mutation_arg_dsyms) == 1:
                namespace_scope = self.safety.namespaces.get(mutated_obj_id, None)
                mutated_obj_aliases = self.safety.aliases.get(mutated_obj_id, None)
                if mutated_obj_aliases is not None:
                    mutated_sym = next(iter(mutated_obj_aliases))
                    mutated_obj = mutated_sym._get_obj()
                    mutation_arg_sym = next(iter(mutation_arg_dsyms))
                    mutation_arg_obj = mutation_arg_sym._get_obj()
                    # TODO: replace int check w/ more general "immutable" check
                    if mutated_sym is not None and mutation_arg_obj is not None and not isinstance(mutation_arg_obj, int):
                        if namespace_scope is None:
                            namespace_scope = NamespaceScope(
                                mutated_obj, self.safety, mutated_sym.name,
                                parent_scope=mutated_sym.containing_scope
                            )
                        namespace_scope.upsert_data_symbol_for_name(
                            len(mutated_obj) - 1, mutation_arg_obj, set(), self.stmt_node,
                            is_subscript=True, overwrite=False, propagate=False
                        )
            # TODO: add mechanism for skipping namespace children in case of list append
            for mutated_sym in self.safety.aliases[mutated_obj_id]:
                mutated_sym.update_deps(mutation_arg_dsyms, overwrite=False, mutated=True)
        if self._contains_lval():
            self._make_lval_data_symbols()
        else:
            if len(self.safety.tracing_manager.saved_store_data) > 0 and self.safety.is_develop:
                logger.warning('saw unexpected state in saved_store_data: %s',
                               self.safety.tracing_manager.saved_store_data)

    def finished_execution_hook(self):
        if self.finished:
            return
        # print('finishing stmt', self.stmt_node)
        self.safety.tracing_manager.seen_stmts.add(self.stmt_id)
        self.handle_dependencies()
        self.safety.tracing_manager.after_stmt_reset_hook()
        self.safety._namespace_gc()
        # self.safety._gc()



from nbsafety.analysis.symbol_edges import get_assignment_lval_and_rval_symbol_refs


def test_simple_implicit_tuple_assignment():
    edges = list(get_assignment_lval_and_rval_symbol_refs('a, b = c, d'))
    assert edges == [('a', 'c'), ('b', 'd')]


def test_unpack_to_one_target_assignment():
    edges = list(get_assignment_lval_and_rval_symbol_refs('a, b = c'))
    assert edges == [('a', 'c'), ('b', 'c')]


def test_simple_explicit_tuple_assignment():
    edges = list(get_assignment_lval_and_rval_symbol_refs('(a, b) = (c, d)'))
    assert edges == [('a', 'c'), ('b', 'd')]


def test_simple_explicit_list_assignment():
    edges = list(get_assignment_lval_and_rval_symbol_refs('[a, b] = [c, d]'))
    assert edges == [('a', 'c'), ('b', 'd')]


def test_multiple_tuple_assignment():
    edges = list(get_assignment_lval_and_rval_symbol_refs('a, b = c = d, e'))
    assert edges == [('a', 'd'), ('b', 'e'), ('c', 'd'), ('c', 'e')]


def test_tuple_assignment_with_dict():
    edges = list(get_assignment_lval_and_rval_symbol_refs('a, b = (d, e), {f: 5, g: h}'))
    assert edges == [('a', 'd'), ('a', 'e'), ('b', 'f'), ('b', 'g'), ('b', None), ('b', 'h')]


def test_assignment_with_constants():
    edges = list(get_assignment_lval_and_rval_symbol_refs('a, b = x + 1, y + 2'))
    assert edges == [('a', 'x'), ('a', None), ('b', 'y'), ('b', None)]


def test_nested_assignment():
    edges = list(get_assignment_lval_and_rval_symbol_refs('a, (b, [c, d]) = (d, e), [(f, g), h]'))
    assert edges == [
        ('a', 'd'),
        ('a', 'e'),
        ('b', 'f'),
        ('b', 'g'),
        ('c', 'h'),
        ('d', 'h'),
    ]



# -*- coding: utf-8 -*-
import logging

from test.utils import make_safety_fixture

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, _safety_state, run_cell_ = make_safety_fixture(setup_cells=['%safety trace_messages enable'])
_safety_fixture, _safety_state, run_cell_ = make_safety_fixture()


def run_cell(cell, **kwargs):
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell, **kwargs)


def test_stack_on_tracing_reenable():
    run_cell("""
x = 42

def fake_func(y):
    z = 9
    return z

new_xs = [fake_func(x) for _ in range(5)]

print(new_xs[0])
""")
